#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2010 by Jason Heeris <jason.heeris@gmail.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
from time import sleep
from collections import deque

try:
    from gi.repository import GObject as gobject
except ImportError:
    import gobject

import os
if "NAUTILUS_PYTHON_REQUIRE_GTK3" in os.environ and os.environ["NAUTILUS_PYTHON_REQUIRE_GTK3"]:
    from gi.repository import Gtk as gtk
else:
    import gtk

from rabbitvcs.vcs import create_vcs_instance, VCS_SVN, VCS_GIT, VCS_DUMMY, VCS_MERCURIAL
from rabbitvcs.util.log import Log
from rabbitvcs import gettext
from rabbitvcs.util.settings import SettingsManager
import rabbitvcs.util.helper

# Yes, * imports are bad. You write it out then.
from contextmenuitems import *

log = Log("rabbitvcs.util.contextmenu")
_ = gettext.gettext

settings = SettingsManager()

class MenuBuilder(object):
    """
    Generalised menu builder class. Subclasses must provide:
    
    connect_to_signal(self, menuitem, callback, callback_args) - connect the
    menu item to a signal (or do whatever needs to be done) so that the callback
    is called upon activation.
    
    In actual fact, a standard GTK compatible method for this is provided by
    this class. All a subclass has to do is define the class parameter "signal",
    and it will be automatically done. 
    
    make_menu_item(self, item, id_magic) - create the menu item for whatever
    toolkit (usually this should be just call a  convenience method on the
    MenuItem instance).
    
    attach_submenu(self, menu_node, submenu_list) - given a list of whatever
    make_menu_item(...) returns, create a submenu and attach it to the given
    node.
    
    top_level_menu(self, items) - in some circumstances we need to treat the top
    level menu differently (eg. Nautilus, because Xenu said so). This processes
    a list of menu items returned by make_menu_item(...) to create the overall
    menu.  
    """
    
    def __init__(self, structure, conditions, callbacks):
        """
        @param  structure: Menu structure
        @type   structure: list
                
        Note on "structure". The menu structure is defined in a list of tuples 
        of two elements each.  The first element is a class - the MenuItem
        subclass that defines the menu interface (see below).
        
        The second element is either None (if there is no submenu) or a list of
        tuples if there is a submenu.  The submenus are generated recursively. 
        FYI, this is a list of tuples so that we retain the desired menu item
        order (dicts do not retain order)
        
            Example:
            [
                (MenuClassOne, [
                    (MenuClassOneSubA, None),
                    (MenuClassOneSubB, None)
                ]),
                (MenuClassTwo, None),
                (MenuClassThree, None)
            ]
            
        """
        # The index is mostly for identifier magic 
        index = 0
        last_level = -1
        last_item = last_menuitem = None        

        stack = [] # ([items], last_item, last_menuitem)
        flat_structure = rabbitvcs.util.helper.walk_tree_depth_first(
                                structure,
                                show_levels=True,
                                preprocess=lambda x: x(conditions, callbacks),
                                filter=lambda x: x.show())
        
        # Here's how this works: we walk the tree, which is a series of (level,
        # MenuItem instance) tuples. We accumulate items in the list in
        # stack[level][0], and when we go back up a level we put them in a
        # submenu (as defined by the subclasses). We need to keep track of the
        # last item on each level, in case they are separators, so that's on the
        # stack too.       
        for (level, item) in flat_structure:
            index += 1

            # Have we dropped back a level? Restore previous context
            if level < last_level:
                # We may have ended up descending several levels (it works, but
                # please no-one write triply nested menus, it's just dumb).
                for num in range(last_level - level):
                    # Remove separators at the end of menus
                    if type(last_item) == MenuSeparator:
                        stack[-1][0].remove(last_menuitem)
                                    
                    (items, last_item, last_menuitem) = stack.pop()
                    
                    # Every time we back out of a level, we attach the list of
                    # items as a submenu, however the subclass wants to do it.                    
                    self.attach_submenu(last_menuitem, items)                                

            # Have we gone up a level? Save the context and create a submenu
            if level > last_level:
                # Skip separators at the start of a menu
                if type(item) == MenuSeparator: continue
                                
                stack.append(([], last_item, last_menuitem))
                
                last_item = last_menuitem = None
        
            # Skip duplicate separators
            if (type(last_item) == type(item) == MenuSeparator and
                level == last_level):
                continue
            
            menuitem = self.make_menu_item(item, index)

            self.connect_signal(menuitem, item.callback, item.callback_args)
        
            stack[-1][0].append(menuitem)

            last_item = item
            last_menuitem = menuitem
            last_level = level

        # Hey, we're out of the loop. Go back up any remaining levels (in case
        # there were submenus at the end) and finish the job.
        for (items, last_item2, last_menuitem2) in stack[:0:-1]:
            
            if type(last_item) == MenuSeparator:
                stack[-1][0].remove(last_menuitem)
            
            self.attach_submenu(last_menuitem2, items)
            
            last_item = last_item2
            last_menuitem = last_menuitem2

        if stack:
            self.menu = self.top_level_menu(stack[0][0])
        else:
            log.debug("Empty top level menu!")
            self.menu = self.top_level_menu([])
            
    def connect_signal(self, menuitem, callback, callback_args):
        if callback:
            if callback_args:
                menuitem.connect(self.signal, callback, callback_args)
            else:
                menuitem.connect(self.signal, callback)

class GtkContextMenu(MenuBuilder):
    """
    Provides a standard Gtk Context Menu class used for all context menus
    in gtk dialogs/windows.
    
    """
    
    signal = "button-press-event"
                            
    def make_menu_item(self, item, id_magic):
        return item.make_gtk_menu_item(id_magic)
    
    def attach_submenu(self, menu_node, submenu_list):
        submenu = gtk.Menu()
        menu_node.set_submenu(submenu)
        [submenu.add(item) for item in submenu_list]
    
    def top_level_menu(self, items):
        menu = gtk.Menu()
        [menu.add(item) for item in items]
        return menu
        
    def show(self, event):        
        self.menu.show_all()
        self.menu.popup(None, None, None, event.button, event.time)

    def get_widget(self):
        return self.menu

class GtkContextMenuCaller:
    """
    Provides an abstract interface to be inherited by dialogs/windows that call
    a GtkContextMenu.  Allows us to have a standard common set of methods we can
    call from the callback object.
    """
    
    def __init__(self):
        pass
    
    def on_context_menu_command_finished(self):
        pass

    def rescan_after_process_exit(self, proc, paths=None):
        self.execute_after_process_exit(proc, self.on_context_menu_command_finished)
        
    def execute_after_process_exit(self, proc, callback=None):
        if callback is None:
            callback = self.on_context_menu_command_finished

        def is_process_still_alive():
            log.debug("is_process_still_alive() for pid: %i" % proc.pid)
            # First we need to see if the process is still running

            retval = proc.poll()
            
            log.debug("%s" % retval)
            
            still_going = (retval is None)

            if not still_going and callable(callback):
                callback()

            return still_going

        # Add our callback function on a 1 second timeout
        gobject.timeout_add_seconds(1, is_process_still_alive)
        
class ContextMenuCallbacks:
    """
    The base class for context menu callbacks.  This is inheritied by
    sub-classes.
    """
    def __init__(self, caller, base_dir, vcs_client, paths=[]):
        """        
        @param  caller: The calling object
        @type   caller: RabbitVCS extension
        
        @param  base_dir: The curent working directory
        @type   base_dir: string

        @param  vcs_client: The vcs client to be used
        @type   vcs_client: rabbitvcs.vcs.create_vcs_instance()
        
        @param  paths: The selected paths
        @type   paths: list
        
        """  
        self.caller = caller
        self.base_dir = base_dir
        self.vcs_client = vcs_client
        self.paths = paths
        
    def debug_shell(self, widget, data1=None, data2=None):
        """
        
        Open up an IPython shell which shares the context of the extension.
        
        See: http://ipython.scipy.org/moin/Cookbook/EmbeddingInGTK
        
        """
        import gtk
        from rabbitvcs.debug.ipython_view import IPythonView
        
        window = gtk.Window()
        window.set_size_request(750,550)
        window.set_resizable(True)
        window.set_position(gtk.WIN_POS_CENTER)
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        ipython_view = IPythonView()
        ipython_view.updateNamespace(locals())
        ipython_view.set_wrap_mode(gtk.WRAP_CHAR)
        ipython_view.show()
        scrolled_window.add(ipython_view)
        scrolled_window.show()
        window.add(scrolled_window)
        window.show()
    
    def refresh_status(self, widget, data1=None, data2=None):
        """
        Refreshes an item status, which is actually just invalidate.
        """
        
        self.debug_invalidate(widget)
    
    def debug_revert(self, widget, data1=None, data2=None):
        client = pysvn.Client()
        for path in self.paths:
            client.revert(path, recurse=True)
        
    def debug_invalidate(self, widget, data1=None, data2=None):
        rabbitvcs_extension = self.caller
        nautilusVFSFile_table = rabbitvcs_extension.nautilusVFSFile_table
        for path in self.paths:
            log.debug("callback_debug_invalidate() called for %s" % path)
            if path in nautilusVFSFile_table:
                nautilusVFSFile_table[path].invalidate_extension_info()
    
    def debug_add_emblem(self, widget, data1=None, data2=None):
        def add_emblem_dialog():
            from subprocess import Popen, PIPE
            command = ["zenity", "--entry", "--title=RabbitVCS", "--text=Emblem to add:"]
            emblem = Popen(command, stdout=PIPE).communicate()[0].replace("\n", "")
            
            rabbitvcs_extension = self.caller
            nautilusVFSFile_table = rabbitvcs_extension.nautilusVFSFile_table
            for path in self.paths:
                if path in nautilusVFSFile_table:
                    nautilusVFSFile_table[path].add_emblem(emblem)
            return False
            
        gobject.idle_add(add_emblem_dialog)
    
    # End debugging callbacks

    def checkout(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("checkout", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def update(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("update", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def commit(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("commit", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def add(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("add", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def check_for_modifications(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("checkmods", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def delete(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("delete", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def revert(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("revert", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def diff(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("diff", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def diff_multiple(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("diff", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def diff_previous_revision(self, widget, data1=None, data2=None):
        guess = self.vcs_client.guess(self.paths[0])
        if guess["vcs"] == rabbitvcs.vcs.VCS_SVN:
            previous_revision_number = self.vcs_client.svn().get_revision(self.paths[0]) - 1
        
            pathrev1 = rabbitvcs.util.helper.create_path_revision_string(self.vcs_client.svn().get_repo_url(self.paths[0]), previous_revision_number)
            pathrev2 = rabbitvcs.util.helper.create_path_revision_string(self.paths[0], "working")
            
            proc = rabbitvcs.util.helper.launch_ui_window("diff", [
                "-s",
                pathrev1,
                pathrev2,
                "--vcs=%s" % rabbitvcs.vcs.VCS_SVN
            ])
            self.caller.rescan_after_process_exit(proc, self.paths)

    def compare_tool(self, widget, data1=None, data2=None):
        pathrev1 = rabbitvcs.util.helper.create_path_revision_string(self.paths[0], "base")
        pathrev2 = rabbitvcs.util.helper.create_path_revision_string(self.paths[0], "working")

        proc = rabbitvcs.util.helper.launch_ui_window("diff", [
            "-s",
            pathrev1,
            pathrev2
        ])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def compare_tool_multiple(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("diff", ["-s"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def compare_tool_previous_revision(self, widget, data1=None, data2=None):
        guess = self.vcs_client.guess(self.paths[0])
        if guess["vcs"] == rabbitvcs.vcs.VCS_SVN:
            previous_revision_number = self.vcs_client.svn().get_revision(self.paths[0]) - 1

            pathrev1 = rabbitvcs.util.helper.create_path_revision_string(self.vcs_client.svn().get_repo_url(self.paths[0]), previous_revision_number)
            pathrev2 = rabbitvcs.util.helper.create_path_revision_string(self.paths[0], "working")

            proc = rabbitvcs.util.helper.launch_ui_window("diff", [
                "-s",
                pathrev1,
                pathrev2,
                "--vcs=%s" % rabbitvcs.vcs.VCS_SVN
            ])
            self.caller.rescan_after_process_exit(proc, self.paths)

    def show_changes(self, widget, data1=None, data2=None):
        pathrev1 = rabbitvcs.util.helper.create_path_revision_string(self.paths[0])
        pathrev2 = pathrev1
        if len(self.paths) == 2:
            pathrev2 = rabbitvcs.util.helper.create_path_revision_string(self.paths[1])
        
        proc = rabbitvcs.util.helper.launch_ui_window("changes", [pathrev1, pathrev2])
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def show_log(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("log", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def rename(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("rename", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def create_patch(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("createpatch", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def apply_patch(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("applypatch", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def properties(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("property_editor", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def about(self, widget, data1=None, data2=None):
        rabbitvcs.util.helper.launch_ui_window("about")
        
    def settings(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("settings", [self.base_dir])
        self.caller.reload_settings(proc)

    def ignore_by_filename(self, widget, data1=None, data2=None):
        path = self.paths[0]
        base_dir = os.path.join(self.base_dir, os.path.dirname(path))
        proc = rabbitvcs.util.helper.launch_ui_window("ignore", [base_dir, os.path.basename(path)])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def ignore_by_file_extension(self, widget, data1=None, data2=None):
        path = self.paths[0]
        pattern = "*%s" % rabbitvcs.util.helper.get_file_extension(path)
        base_dir = os.path.join(self.base_dir, os.path.dirname(path))
        proc = rabbitvcs.util.helper.launch_ui_window("ignore", [base_dir, pattern])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def get_lock(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("lock", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def branch_tag(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("branch", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def switch(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("switch", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def merge(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("merge", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def _import(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("import", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def export(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("export", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def svn_export(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("export", ["--vcs=svn", self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def git_export(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("export", ["--vcs=git", self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def update_to_revision(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("updateto", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def mark_resolved(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("markresolved", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
        
    def annotate(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("annotate", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def unlock(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("unlock", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
        
    def create_repository(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("create", ["--vcs", "svn", self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def relocate(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("relocate", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def cleanup(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("cleanup", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def restore(self, widget, data1=None, data2=None):
        guess = self.vcs_client.guess(self.paths[0])
        if guess["vcs"] == rabbitvcs.vcs.VCS_SVN:
            proc = rabbitvcs.util.helper.launch_ui_window("update", self.paths)
        elif guess["vcs"] == rabbitvcs.vcs.VCS_GIT:
            proc = rabbitvcs.util.helper.launch_ui_window("checkout", ["-q", "--vcs", "git"] + self.paths)

        self.caller.rescan_after_process_exit(proc, self.paths)

    def _open(self, widget, data1=None, data2=None):
        pass
    
    def browse_to(self, widget, data1=None, data2=None):
        pass

    def repo_browser(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("browser", [self.paths[0]])

    def initialize_repository(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("create", ["--vcs", "git", self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def clone(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("clone", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def push(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("push", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def branches(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("branches", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def tags(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("tags", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def remotes(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("remotes", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def clean(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("clean", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def reset(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("reset", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def stage(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("stage", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def unstage(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("unstage", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, self.paths)

    def edit_conflicts(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("editconflicts", [self.paths[0]])
        self.caller.rescan_after_process_exit(proc, [self.paths[0]])

class ContextMenuConditions:
    """
    Provides a standard interface to checking conditions for menu items.
    
    This class should never be instantied directly, rather the narrowly defined
    FileManagerContextMenuConditions and GtkFilesContextMenuConditions classes
    should be called.
    
    """    
    def __init__(self):
        pass

    def generate_path_dict(self, paths):
        self.path_dict = {
            "length": len(paths)
        }

        checks = {
            "is_svn"                        : lambda path: (self.vcs_client.guess(path)["vcs"] == VCS_SVN),
            "is_git"                        : lambda path: (self.vcs_client.guess(path)["vcs"] == VCS_GIT),
            "is_mercurial"                  : lambda path: (self.vcs_client.guess(path)["vcs"] == VCS_MERCURIAL),
            "is_dir"                        : os.path.isdir,
            "is_file"                       : os.path.isfile,
            "exists"                        : os.path.exists,
            "is_working_copy"               : self.vcs_client.is_working_copy,
            "is_in_a_or_a_working_copy"     : self.vcs_client.is_in_a_or_a_working_copy,
            "is_versioned"                  : self.vcs_client.is_versioned,
            "is_normal"                     : lambda path: self.statuses[path].simple_content_status() == "unchanged" and self.statuses[path].simple_metadata_status() == "normal",
            "is_added"                      : lambda path: self.statuses[path].simple_content_status() == "added",
            "is_modified"                   : lambda path: self.statuses[path].simple_content_status() == "modified" or self.statuses[path].simple_metadata_status() == "modified",
            "is_deleted"                    : lambda path: self.statuses[path].simple_content_status() == "deleted",
            "is_ignored"                    : lambda path: self.statuses[path].simple_content_status() == "ignored",
            "is_locked"                     : self.vcs_client.is_locked,
            "is_missing"                    : lambda path: self.statuses[path].simple_content_status() == "missing",
            "is_conflicted"                 : lambda path: self.statuses[path].simple_content_status() == "complicated",
            "is_obstructed"                 : lambda path: self.statuses[path].simple_content_status() == "obstructed",
            "has_unversioned"               : lambda path: "unversioned" in self.text_statuses,
            "has_added"                     : lambda path: "added" in self.text_statuses,
            "has_modified"                  : lambda path: "modified" in self.text_statuses or "modified" in self.prop_statuses,
            "has_deleted"                   : lambda path: "deleted" in self.text_statuses,
            "has_ignored"                   : lambda path: "ignored" in self.text_statuses,
            "has_missing"                   : lambda path: "missing" in self.text_statuses,
            "has_conflicted"                : lambda path: "complicated" in self.text_statuses,
            "has_obstructed"                : lambda path: "obstructed" in self.text_statuses
        }
        
        for key,func in checks.items():
            self.path_dict[key] = False

        # Each path gets tested for each check
        # If a check has returned True for any path, skip it for remaining paths
        for path in paths:
            for key, func in checks.items():
                try:
                    self.path_dict[key] = func(path)
                except KeyError, e:
                    self.path_dict[key] = False

    def checkout(self, data=None):
        if self.path_dict["length"] == 1:
            if self.path_dict["is_git"]:
                return (self.path_dict["is_in_a_or_a_working_copy"] and
                    self.path_dict["is_versioned"])
            else:
                return (self.path_dict["is_dir"] and
                        not self.path_dict["is_working_copy"])
    
        return False
    
    def update(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                not self.path_dict["is_added"])
                        
    def commit(self, data=None):
        if self.path_dict["is_svn"] or self.path_dict["is_git"] or self.path_dict["is_mercurial"]:
            if self.path_dict["is_in_a_or_a_working_copy"]:
                if (self.path_dict["is_added"] or
                        self.path_dict["is_modified"] or
                        self.path_dict["is_deleted"] or
                        not self.path_dict["is_versioned"]):
                    return True
                elif (self.path_dict["is_dir"]):
                    return True
        return False

    def diff_menu(self, data=None):
        return self.path_dict["is_in_a_or_a_working_copy"]
    
    def diff_multiple(self, data=None):
        if (self.path_dict["length"] == 2 and 
                self.path_dict["is_versioned"] and
                self.path_dict["is_in_a_or_a_working_copy"]):
            return True
        return False

    def compare_tool_multiple(self, data=None):
        if (self.path_dict["length"] == 2 and 
                self.path_dict["is_versioned"] and
                self.path_dict["is_in_a_or_a_working_copy"]):
            return True
        return False
        
    def diff(self, data=None):
        if (self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"] and
                (self.path_dict["is_modified"] or self.path_dict["has_modified"] or 
                self.path_dict["is_conflicted"] or self.path_dict["has_conflicted"])):
            return True        
        return False

    def diff_previous_revision(self, data=None):
        if (self.path_dict["is_svn"] and
                self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"]):
            return True        
        return False

    def compare_tool(self, data=None):
        if (self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"] and
                (self.path_dict["is_modified"] or self.path_dict["has_modified"] or 
                self.path_dict["is_conflicted"] or self.path_dict["has_conflicted"])):
            return True        
        return False

    def compare_tool_previous_revision(self, data=None):
        if (self.path_dict["is_svn"] and
                self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"]):
            return True        
        return False

    def show_changes(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
            self.path_dict["is_versioned"] and 
            self.path_dict["length"] in (1,2))
        
    def show_log(self, data=None):
        return (self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                not self.path_dict["is_added"])
        
    def add(self, data=None):
        if not self.path_dict["is_svn"]:
            return False
            
        if (self.path_dict["is_dir"] and
                self.path_dict["is_in_a_or_a_working_copy"]):
            return True
        elif (not self.path_dict["is_dir"] and
                self.path_dict["is_in_a_or_a_working_copy"] and
                not self.path_dict["is_versioned"]):
            return True
        return False

    def check_for_modifications(self, data=None):
        return (self.path_dict["is_working_copy"] or
            self.path_dict["is_versioned"])

    def rename(self, data=None):
        return (self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"])
        
    def delete(self, data=None):
        return (self.path_dict["exists"] or self.path_dict["is_versioned"]) and \
            not self.path_dict["is_deleted"]
        
    def revert(self, data=None):
        if self.path_dict["is_in_a_or_a_working_copy"]:
            if (self.path_dict["is_added"] or
                    self.path_dict["is_modified"] or
                    self.path_dict["is_deleted"]):
                return True
            else:
                if (self.path_dict["is_dir"] and
                        (self.path_dict["has_added"] or
                        self.path_dict["has_modified"] or
                        self.path_dict["has_deleted"] or
                        self.path_dict["has_missing"])):
                    return True
        return False
        
    def annotate(self, data=None):
        return (self.path_dict["length"] == 1 and
                not self.path_dict["is_dir"] and
                self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                not self.path_dict["is_added"])
        
    def properties(self, data=None):
        return (self.path_dict["length"] == 1 and
                self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"])

    def create_patch(self, data=None):
        if self.path_dict["is_in_a_or_a_working_copy"]:
            if (self.path_dict["is_added"] or
                    self.path_dict["is_modified"] or
                    self.path_dict["is_deleted"] or
                    not self.path_dict["is_versioned"]):
                return True
            elif (self.path_dict["is_dir"] and
                    (self.path_dict["has_added"] or
                    self.path_dict["has_modified"] or
                    self.path_dict["has_deleted"] or
                    self.path_dict["has_unversioned"] or
                    self.path_dict["has_missing"])):
                return True
        return False
    
    def apply_patch(self, data=None):
        if self.path_dict["is_in_a_or_a_working_copy"]:
            return True
        return False
    
    def add_to_ignore_list(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                not self.path_dict["is_versioned"])

    def ignore_by_filename(self, *args):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                not self.path_dict["is_versioned"])
    
    def ignore_by_file_extension(self, *args):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                not self.path_dict["is_versioned"])

    def refresh_status(self, data=None):
        return True

    def get_lock(self, data=None):
        return self.path_dict["is_versioned"]

    def branch_tag(self, data=None):
        return self.path_dict["is_versioned"]

    def relocate(self, data=None):
        return self.path_dict["is_versioned"]

    def switch(self, data=None):
        return self.path_dict["is_versioned"]

    def merge(self, data=None):
        return self.path_dict["is_versioned"]

    def _import(self, data=None):
        return (self.path_dict["length"] == 1 and
                not self.path_dict["is_in_a_or_a_working_copy"])

    def export(self, data=None):
        return (self.path_dict["length"] == 1)

    def svn_export(self, data=None):
        return self.export(data)

    def git_export(self, data=None):
        return self.export(data)

    def update_to_revision(self, data=None):
        return (self.path_dict["length"] == 1 and
                self.path_dict["is_versioned"] and 
                self.path_dict["is_in_a_or_a_working_copy"])
    
    def mark_resolved(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                self.path_dict["is_conflicted"])
            
    def create_repository(self, data=None):
        return (self.path_dict["length"] == 1 and
                not self.path_dict["is_in_a_or_a_working_copy"])

    def unlock(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                (self.path_dict["is_dir"] or self.path_dict["is_locked"]))

    def cleanup(self, data=None):
        return self.path_dict["is_versioned"]

    def browse_to(self, data=None):
        return self.path_dict["exists"]
    
    def _open(self, data=None):
        return self.path_dict["is_file"]
    
    def restore(self, data=None):
        return self.path_dict["has_missing"]

    def update(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                not self.path_dict["is_added"])

    def repo_browser(self, data=None):
        return True

    def rabbitvcs(self, data=None):
        return False

    def rabbitvcs_svn(self, data=None):
        return (self.path_dict["is_svn"] or 
            not self.path_dict["is_in_a_or_a_working_copy"])

    def rabbitvcs_git(self, data=None):
        return (self.path_dict["is_git"] or 
            not self.path_dict["is_in_a_or_a_working_copy"])

    def rabbitvcs_mercurial(self, data=None):
        return (self.path_dict["is_mercurial"] or 
            not self.path_dict["is_in_a_or_a_working_copy"])
    
    def debug(self, data=None):
        return settings.get("general", "show_debug")
    
    def separator(self, data=None):
        return True
    
    def help(self, data=None):
        return False
    
    def settings(self, data=None):
        return True
    
    def about(self, data=None):
        return True
    
    def bugs(self, data=None):
        return True

    def initialize_repository(self, data=None):
        return (self.path_dict["is_dir"] and
            not self.path_dict["is_in_a_or_a_working_copy"])

    def clone(self, data=None):
        return (self.path_dict["is_dir"] and
            not self.path_dict["is_in_a_or_a_working_copy"])
            
    def push(self, data=None):
        return (self.path_dict["is_git"] or self.path_dict["is_mercurial"])

    def branches(self, data=None):
        return (self.path_dict["is_git"])

    def tags(self, data=None):
        return (self.path_dict["is_git"])
        
    def remotes(self, data=None):
        return (self.path_dict["is_git"])

    def clean(self, data=None):
        return (self.path_dict["is_git"])

    def reset(self, data=None):
        return (self.path_dict["is_git"])

    def stage(self, data=None):
        if self.path_dict["is_git"]:
            if (self.path_dict["is_dir"] and
                    self.path_dict["is_in_a_or_a_working_copy"]):
                return True
            elif (not self.path_dict["is_dir"] and
                    self.path_dict["is_in_a_or_a_working_copy"] and
                    not self.path_dict["is_versioned"]):
                return True
        return False

    def unstage(self, data=None):
        if self.path_dict["is_git"]:
            if (self.path_dict["is_dir"] and
                    self.path_dict["is_in_a_or_a_working_copy"]):
                return True
            elif (not self.path_dict["is_dir"] and
                    self.path_dict["is_in_a_or_a_working_copy"] and
                    self.path_dict["is_added"]):
                return True
        return False

    def edit_conflicts(self, data=None):
        return (self.path_dict["is_in_a_or_a_working_copy"] and
                self.path_dict["is_versioned"] and
                self.path_dict["is_conflicted"])

class GtkFilesContextMenuCallbacks(ContextMenuCallbacks):
    """
    A callback class created for GtkFilesContextMenus.  This class inherits from
    the standard ContextMenuCallbacks class and overrides some methods.
    """
    def __init__(self, caller, base_dir, vcs_client, paths=[]):
        """        
        @param  caller: The calling object
        @type   caller: RabbitVCS extension
        
        @param  base_dir: The curent working directory
        @type   base_dir: string

        @param  vcs_client: The vcs client to be used
        @type   vcs_client: rabbitvcs.vcs.create_vcs_instance()
        
        @param  paths: The selected paths
        @type   paths: list
        
        """  
        ContextMenuCallbacks.__init__(self, caller, base_dir, vcs_client, paths)

    def _open(self, widget, data1=None, data2=None):
        for path in self.paths:
            rabbitvcs.util.helper.open_item(path)

    def add(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("add", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def revert(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("revert", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def mark_resolved(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("markresolved", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def browse_to(self, widget, data1=None, data2=None):
        rabbitvcs.util.helper.browse_to_item(self.paths[0])

    def delete(self, widget, data1=None, data2=None):
        if len(self.paths) > 0:
            proc = rabbitvcs.util.helper.launch_ui_window("delete", self.paths)
            self.caller.rescan_after_process_exit(proc, self.paths)
  
    def update(self, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("update", self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
    
    def unlock(self, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("unlock", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)
            
    def show_log(self, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("log", self.paths)
        self.caller.rescan_after_process_exit(proc, [self.paths[0]])

    def stage(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("stage", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

    def unstage(self, widget, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window("unstage", ["-q"] + self.paths)
        self.caller.rescan_after_process_exit(proc, self.paths)

class GtkFilesContextMenuConditions(ContextMenuConditions):
    """
    Sub-class for ContextMenuConditions for our dialogs.  Allows us to override 
    some generic condition methods with condition logic more suitable 
    to the dialogs.
    
    """
    def __init__(self, vcs_client, paths=[]):
        """    
        @param  vcs_client: The vcs client to be used
        @type   vcs_client: rabbitvcs.vcs.create_vcs_instance()
        
        @param  paths: The selected paths
        @type   paths: list
        
        """
        self.vcs_client = vcs_client
        self.paths = paths
        self.statuses = {}

        self.generate_statuses(self.paths)
        self.generate_path_dict(self.paths)

    def generate_statuses(self, paths):
        self.statuses = {}
        for path in paths:
            if not path:
                continue
                
            statuses_tmp = self.vcs_client.statuses(path, invalidate=True)
            for status in statuses_tmp:
                self.statuses[status.path] = status

        self.text_statuses = [self.statuses[key].simple_content_status() for key in self.statuses.keys()]
        self.prop_statuses = [self.statuses[key].simple_metadata_status() for key in self.statuses.keys()]
        
class GtkFilesContextMenu:
    """
    Defines context menu items for a table with files
    
    """
    def __init__(self, caller, event, base_dir, paths=[], 
            conditions=None, callbacks=None):
        """    
        @param  caller: The calling object
        @type   caller: RabbitVCS extension
        
        @param  base_dir: The curent working directory
        @type   base_dir: string
        
        @param  paths: The selected paths
        @type   paths: list
        
        @param  conditions: The conditions class that determines menu item visibility
        @kind   conditions: ContextMenuConditions
        
        @param  callbacks: The callbacks class that determines what actions are taken
        @kind   callbacks: ContextMenuCallbacks
        
        """        
        self.caller = caller
        self.event = event
        self.paths = paths
        self.base_dir = base_dir
        self.vcs_client = create_vcs_instance()

        self.conditions = conditions
        if self.conditions is None:
            self.conditions = GtkFilesContextMenuConditions(self.vcs_client, paths)

        self.callbacks = callbacks
        if self.callbacks is None:
            self.callbacks = GtkFilesContextMenuCallbacks(
                self.caller, 
                self.base_dir,
                self.vcs_client, 
                paths
            )
            
        ignore_items = get_ignore_list_items(paths)

        # The first element of each tuple is a key that matches a
        # ContextMenuItems item.  The second element is either None when there
        # is no submenu, or a recursive list of tuples for desired submenus.
        self.structure = [
            (MenuDiff, None),
            (MenuCompareTool, None),
            (MenuUnlock, None),
            (MenuShowLog, None),
            (MenuOpen, None),
            (MenuBrowseTo, None),
            (MenuDelete, None),
            (MenuRevert, None),
            (MenuRestore, None),
            (MenuEditConflicts, None),
            (MenuMarkResolved, None),
            (MenuCreatePatch, None),
            (MenuAdd, None),
            (MenuStage, None),
            (MenuUnstage, None),
            (MenuAddToIgnoreList, ignore_items)
        ]
        
    def show(self):
        if len(self.paths) == 0:
            return

        context_menu = GtkContextMenu(self.structure, self.conditions, self.callbacks)
        context_menu.show(self.event)

    def get_menu(self):
        context_menu = GtkContextMenu(self.structure, self.conditions, self.callbacks)
        return context_menu.menu

class MainContextMenuCallbacks(ContextMenuCallbacks):
    """
    The callback class used for the main context menu.  This inherits from
    and overrides the ContextMenuCallbacks class.
    
    """   
    def __init__(self, caller, base_dir, vcs_client, paths=[]):
        """        
        @param  caller: The calling object
        @type   caller: RabbitVCS extension
        
        @param  base_dir: The curent working directory
        @type   base_dir: string

        @param  vcs_client: The vcs client to be used
        @type   vcs_client: rabbitvcs.vcs.create_vcs_instance()
        
        @param  paths: The selected paths
        @type   paths: list
        
        """  
        ContextMenuCallbacks.__init__(self, caller, base_dir, vcs_client, paths)

class MainContextMenuConditions(ContextMenuConditions):
    """
    Sub-class for ContextMenuConditions used for file manager extensions.  
    Allows us to override some generic condition methods with condition logic 
    more suitable to the dialogs.
    
    """
    def __init__(self, vcs_client, paths=[]):
        """    
        @param  vcs_client: The vcs client to be used
        @type   vcs_client: rabbitvcs.vcs.create_vcs_instance()
        
        @param  paths: The selected paths
        @type   paths: list
        
        """   

        self.vcs_client = vcs_client
        self.paths = paths
        self.statuses = {}
        
        self.generate_statuses(paths)
        self.generate_path_dict(paths)
        
    # FIXME: major bottleneck
    def generate_statuses(self, paths):
        self.statuses = {}
        for path in paths:
            if not path:
                continue
                
            statuses_tmp = self.vcs_client.statuses(path)
            for status in statuses_tmp:
                self.statuses[status.path] = status

        self.text_statuses = [self.statuses[key].simple_content_status() for key in self.statuses.keys()]
        self.prop_statuses = [self.statuses[key].simple_metadata_status() for key in self.statuses.keys()]

class MainContextMenu:
    """
    Defines and composes the main context menu.

    """
    def __init__(self, caller, base_dir, paths=[], 
            conditions=None, callbacks=None):
        """    
        @param  caller: The calling object
        @type   caller: RabbitVCS extension
        
        @param  base_dir: The curent working directory
        @type   base_dir: string
        
        @param  paths: The selected paths
        @type   paths: list
        
        @param  conditions: The conditions class that determines menu item visibility
        @kind   conditions: ContextMenuConditions
        
        @param  callbacks: The callbacks class that determines what actions are taken
        @kind   callbacks: ContextMenuCallbacks
        
        """        
        self.caller = caller
        self.paths = paths
        self.base_dir = base_dir
        self.vcs_client = create_vcs_instance()

        self.conditions = conditions
        if self.conditions is None:
            self.conditions = MainContextMenuConditions(self.vcs_client, paths)

        self.callbacks = callbacks
        if self.callbacks is None:
            self.callbacks = MainContextMenuCallbacks(
                self.caller, 
                self.base_dir,
                self.vcs_client, 
                paths
            )
            
        ignore_items = get_ignore_list_items(paths)

        # The first element of each tuple is a key that matches a
        # ContextMenuItems item.  The second element is either None when there
        # is no submenu, or a recursive list of tuples for desired submenus.        
        self.structure = [
            (MenuDebug, [
                (MenuBugs, None),
                (MenuDebugShell, None),
                (MenuRefreshStatus, None),
                (MenuDebugRevert, None),
                (MenuDebugInvalidate, None),
                (MenuDebugAddEmblem, None)
            ]),
            (MenuUpdate, None),
            (MenuCommit, None),
            (MenuPush, None),
            (MenuRabbitVCSSvn, [
                (MenuCheckout, None),
                (MenuDiffMenu, [
                    (MenuDiff, None),
                    (MenuDiffPrevRev, None),
                    (MenuDiffMultiple, None),
                    (MenuCompareTool, None),
                    (MenuCompareToolPrevRev, None),
                    (MenuCompareToolMultiple, None),
                    (MenuShowChanges, None),
                ]),
                (MenuShowLog, None),
                (MenuRepoBrowser, None),
                (MenuCheckForModifications, None),
                (MenuSeparator, None),
                (MenuAdd, None),
                (MenuAddToIgnoreList, ignore_items),
                (MenuSeparator, None),
                (MenuUpdateToRevision, None),
                (MenuRename, None),
                (MenuDelete, None),
                (MenuRevert, None),
                (MenuEditConflicts, None),
                (MenuMarkResolved, None),
                (MenuRelocate, None),
                (MenuGetLock, None),
                (MenuUnlock, None),
                (MenuCleanup, None),
                (MenuSeparator, None),
                (MenuSVNExport, None),
                (MenuCreateRepository, None),
                (MenuImport, None),
                (MenuSeparator, None),
                (MenuBranchTag, None),
                (MenuSwitch, None),
                (MenuMerge, None),
                (MenuSeparator, None),
                (MenuAnnotate, None),
                (MenuSeparator, None),
                (MenuCreatePatch, None),
                (MenuApplyPatch, None),
                (MenuProperties, None),
                (MenuSeparator, None),
                (MenuSettings, None),
                (MenuAbout, None)
            ]),
            (MenuRabbitVCSGit, [
                (MenuClone, None),
                (MenuInitializeRepository, None),
                (MenuSeparator, None),
                (MenuDiffMenu, [
                    (MenuDiff, None),
                    (MenuDiffPrevRev, None),
                    (MenuDiffMultiple, None),
                    (MenuCompareTool, None),
                    (MenuCompareToolPrevRev, None),
                    (MenuCompareToolMultiple, None),
                    (MenuShowChanges, None),
                ]),
                (MenuShowLog, None),
                (MenuStage, None),
                (MenuUnstage, None),
                (MenuAddToIgnoreList, ignore_items),
                (MenuSeparator, None),
                (MenuRename, None),
                (MenuDelete, None),
                (MenuRevert, None),
                (MenuClean, None),
                (MenuReset, None),
                (MenuCheckout, None),
                (MenuSeparator, None),
                (MenuBranches, None),
                (MenuTags, None),
                (MenuRemotes, None),
                (MenuSeparator, None),
                (MenuGitExport, None),
                (MenuMerge, None),
                (MenuSeparator, None),
                (MenuAnnotate, None),
                (MenuSeparator, None),
                (MenuCreatePatch, None),
                (MenuApplyPatch, None),
                (MenuSeparator, None),
                (MenuSettings, None),
                (MenuAbout, None)
            ]),
            (MenuRabbitVCSMercurial, [
                (MenuSettings, None),
                (MenuAbout, None)
            ])
        ]
        
    def get_menu(self):
        pass

def TestMenuItemFunctions():
    """
    This is a test for developers to ensure that they've written all the
    necessary conditions and callbacks (and haven't made any typos).
    
    What it does:
      - build a list of all the subclasses of MenuItem
      - build lists of the methods in ContextMenuConditions/ContextMenuCallbacks
      - checks to see whether all the MenuItems conditions and callbacks have
        been assigned - if not, a message is printed 
    """
    
    # These are some simple tests:
    import inspect
    import types
    
    import contextmenuitems
    
    menu_item_subclasses = []
    
    # Let's create a list of all MenuItem subclasses
    for name in dir(contextmenuitems):
        entity = getattr(contextmenuitems, name)
        if type(entity) == types.ClassType:
            mro = inspect.getmro(entity)
            if (entity is not contextmenuitems.MenuItem and
                contextmenuitems.MenuItem in mro):
                menu_item_subclasses.append(entity)
    
    condition_functions = []
    
    # Now let's create a list of all of the functions in our conditions class
    for name in dir(ContextMenuConditions):
        entity = getattr(ContextMenuConditions, name)
        if type(entity) == types.UnboundMethodType and not name.startswith("__"):
            condition_functions.append(entity)
    
    callback_functions = []
    
    # ...and in our callbacks class
    for name in dir(ContextMenuCallbacks):
        entity = getattr(ContextMenuCallbacks, name)
        if type(entity) == types.UnboundMethodType and not name.startswith("__"):
            condition_functions.append(entity)
    
    for cls in menu_item_subclasses:
        item = cls(ContextMenuConditions(), ContextMenuCallbacks(None, None, None, None))
        if not item.found_condition:
            print "Did not find condition function in ContextMenuConditions " \
                  "for %s (type: %s)" % (item.identifier, cls)
        if not item.callback:
            print "Did not find callback function in ContextMenuCallbacks " \
                  "for %s (type: %s)" % (item.identifier, cls)
            

if __name__ == "__main__":
    TestMenuItemFunctions()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os

def splitall(path):
    """Split a path into all of its parts.

    From: Python Cookbook, Credit: Trent Mick
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def relativepath(fromdir, tofile):
    """Find relative path from 'fromdir' to 'tofile'.

    An absolute path is returned if 'fromdir' and 'tofile'
    are on different drives. Martin Bless, 2004-03-22.
    """
    f1name = os.path.abspath(tofile)
    if os.path.splitdrive(f1name)[0]:
        hasdrive = True
    else:
        hasdrive = False
    f1basename = os.path.basename(tofile)
    f1dirname = os.path.dirname(f1name)
    f2dirname = os.path.abspath(fromdir)
    f1parts = splitall(f1dirname)
    f2parts = splitall(f2dirname)
    if hasdrive and (f1parts[0].lower() <> f2parts[0].lower()):
        "Return absolute path since we are on different drives."
        return f1name
    while f1parts and f2parts:
        if hasdrive:
            if f1parts[0].lower() <> f2parts[0].lower():
                break
        else:
            if f1parts[0] <> f2parts[0]:
                break
        del f1parts[0]
        del f2parts[0]
    result = ['..' for part in f2parts]
    result.extend(f1parts)
    result.append(f1basename)
    return os.sep.join(result)



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2010 by Jason Heeris <jason.heeris@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path

import os
if "NAUTILUS_PYTHON_REQUIRE_GTK3" in os.environ and os.environ["NAUTILUS_PYTHON_REQUIRE_GTK3"]:
    from gi.repository import Gtk as gtk
    GTK3 = True
    ICON_SIZE_BUTTON = gtk.IconSize.BUTTON
    ICON_SIZE_DIALOG = gtk.IconSize.DIALOG
else:
    import gtk
    GTK3 = False
    ICON_SIZE_BUTTON = gtk.ICON_SIZE_BUTTON
    ICON_SIZE_DIALOG = gtk.ICON_SIZE_DIALOG
    
from collections import defaultdict
import rabbitvcs.ui
import rabbitvcs.ui.widget

import rabbitvcs.vcs
from rabbitvcs.services.checkerservice import StatusCheckerStub as StatusChecker
from rabbitvcs.ui import STATUS_EMBLEMS

from rabbitvcs.util.log import Log
log = Log("rabbitvcs.ui.property_page")

from rabbitvcs import gettext
_ = gettext.gettext

class PropertyPage(rabbitvcs.ui.GtkBuilderWidgetWrapper):
    
    gtkbuilder_filename = "property_page"
    gtkbuilder_id = "prop_page_scroller"
    
    def __init__(self, paths, vcs=None, claim_domain=True):
        rabbitvcs.ui.GtkBuilderWidgetWrapper.__init__(self,
                                                 claim_domain=claim_domain)
        self.paths = paths
        self.vcs = vcs or rabbitvcs.vcs.VCS()
        
        self.info_pane = self.get_widget("property_page") 
        
        if len(paths) == 1:
            file_info = FileInfoPane(paths[0], self.vcs,
                                     claim_domain=self.claim_domain)
            self.info_pane.pack_start(file_info.get_widget(),
                                        expand=False,
                                        fill=False,
                                        padding=0)
        elif len(paths) > 1:
            try:
                for path in paths:
                    expander = FileInfoExpander(path, self.vcs,
                                                claim_domain=self.claim_domain)
                    self.info_pane.pack_start(expander.get_widget(),
                                        expand=False,
                                        fill=False,
                                        padding=0)
            except Exception, ex:
                log.exception(ex)
                raise
                   
class FileInfoPane(rabbitvcs.ui.GtkBuilderWidgetWrapper):

    gtkbuilder_filename = "property_page"
    gtkbuilder_id = "file_info_table"
    
    def __init__(self, path, vcs=None, claim_domain=True):
        rabbitvcs.ui.GtkBuilderWidgetWrapper.__init__(self,
                                                 claim_domain=claim_domain)
        
        self.path = path
        self.vcs = vcs or rabbitvcs.vcs.VCS()
        self.checker = StatusChecker() 
               
        self.get_widget("file_name").set_text(os.path.basename(path))
        
        self.status = self.checker.check_status(path,
                                                recurse = False, 
                                                invalidate = False,
                                                summary = False)

        self.get_widget("vcs_type").set_text(self.status.vcs_type)
        
        self.get_widget("content_status").set_text(unicode(self.status.simple_content_status()))
        self.get_widget("prop_status").set_text(unicode(self.status.simple_metadata_status()))
        

        self.set_icon_from_status(self.get_widget("content_status_icon"),
                                                  self.status.simple_content_status())

        self.set_icon_from_status(self.get_widget("prop_status_icon"),
                                                  self.status.simple_metadata_status())

        

        self.set_icon_from_status(self.get_widget("vcs_icon"),
                                  self.status.single, ICON_SIZE_DIALOG)
        
        additional_props_table = rabbitvcs.ui.widget.KeyValueTable(
                                    self.get_additional_info())

        additional_props_table.show()

        self.get_widget("file_info_table").pack_start(additional_props_table,
                                                        expand=False,
                                                        fill=False,
                                                        padding=0)
                        
    def set_icon_from_status(self, icon, status, size=ICON_SIZE_BUTTON):
        if status in rabbitvcs.ui.STATUS_EMBLEMS:
            icon.set_from_icon_name("emblem-" + STATUS_EMBLEMS[status], size)

    def get_additional_info(self):
        vcs_type = rabbitvcs.vcs.guess_vcs(self.path)['vcs']
                
        if(vcs_type == rabbitvcs.vcs.VCS_SVN):
            return self.get_additional_info_svn()
        else:
            return []
                
    def get_additional_info_svn(self):
        
        repo_url = self.vcs.svn().get_repo_url(self.path)
        
        return [
            (_("Repository URL"), repo_url)]
        

class FileInfoExpander(rabbitvcs.ui.GtkBuilderWidgetWrapper):

    gtkbuilder_filename = "property_page"
    gtkbuilder_id = "file_info_expander"

    def __init__(self, path, vcs=None, claim_domain=True):
        
        # Might be None, but that's okay, only subclasses use it
        self.vcs = vcs
        
        rabbitvcs.ui.GtkBuilderWidgetWrapper.__init__(self,
                                                 claim_domain=claim_domain)
        self.path = path
        self.get_widget("file_expander_path").set_label(path)
        
        # Do a lazy evaluate for this
        self.file_info = None
        
        self.expander = self.get_widget()
        
        # There seems to be no easy way to connect to this in gtkbuilder
        self.expander.connect("notify::expanded", self.on_expand)

    def on_expand(self, param_spec, user_data):
        if self.expander.get_expanded() and not self.file_info:
            self.file_info = FileInfoPane(self.path, self.vcs,
                                          claim_domain=self.claim_domain
                                          ).get_widget()
            self.expander.add(self.file_info)

class PropertyPageLabel(rabbitvcs.ui.GtkBuilderWidgetWrapper):
    gtkbuilder_filename = "property_page"
    gtkbuilder_id = "property_page_label"    



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceNonView
from rabbitvcs.ui.action import SVNAction
import rabbitvcs.vcs

from rabbitvcs import gettext
_ = gettext.gettext

class SVNCleanup(InterfaceNonView):
    """
    This class provides a handler to the Cleanup window view.
    The idea is that it displays a large folder icon with a label like
    "Please Wait...".  Then when it finishes cleaning up, the label will 
    change to "Finished cleaning up /path/to/folder"
    
    """

    def __init__(self, path):
        InterfaceNonView.__init__(self)
        self.path = path
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()

    def start(self):
        self.action = SVNAction(
            self.svn,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        
        self.action.append(self.action.set_header, _("Cleanup"))
        self.action.append(self.action.set_status, _("Cleaning Up..."))
        self.action.append(self.svn.cleanup, self.path)
        self.action.append(self.action.set_status, _("Completed Cleanup"))
        self.action.append(self.action.finish)
        self.action.start()

        
if __name__ == "__main__":
    from rabbitvcs.ui import main
    (options, paths) = main(usage="Usage: rabbitvcs cleanup [path]")
            
    window = SVNCleanup(paths[0])
    window.register_gtk_quit()
    window.start()
    gtk.main()



#!/usr/bin/python

from sys import argv
import os
import subprocess

def cleanup(modules):
    for module in modules:
        subprocess.call(["python", module, "--cleanup"])

modules = [
    "branch.py", 
    "stage.py",
    "config.py",
    "commit.py",
    "tag.py",
    "remove.py",
    "clone.py",
    "move.py",
    "pull.py",
    "remote.py"
]

if len(argv) == 2 and  argv[1] == "--cleanup":
    cleanup(modules)

for module in modules:
    if subprocess.call(["python", module]) == 1:
        raise SystemExit("Module test failed")

cleanup(modules)



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os
import thread

import pygtk
import gobject
import gtk
from time import sleep

from rabbitvcs.ui import InterfaceView
from rabbitvcs.util.contextmenu import GtkFilesContextMenu, GtkContextMenuCaller
import rabbitvcs.ui.action
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.util
import rabbitvcs.util.helper
from rabbitvcs.util.log import Log
from rabbitvcs.util.decorators import gtk_unsafe
import rabbitvcs.vcs.status

log = Log("rabbitvcs.ui.commit")

from rabbitvcs import gettext
_ = gettext.gettext

gtk.gdk.threads_init()

class Commit(InterfaceView, GtkContextMenuCaller):
    """
    Provides a user interface for the user to commit working copy
    changes to a repository.  Pass it a list of local paths to commit.
    
    """
    SETTINGS = rabbitvcs.util.settings.SettingsManager()

    TOGGLE_ALL = False
    SHOW_UNVERSIONED = SETTINGS.get("general", "show_unversioned_files")

    # This keeps track of any changes that the user has made to the row
    # selections
    changes = {}

    def __init__(self, paths, base_dir=None, message=None):
        """
        
        @type  paths:   list of strings
        @param paths:   A list of local paths.
        
        """
        InterfaceView.__init__(self, "commit", "Commit")

        self.base_dir = base_dir
        self.vcs = rabbitvcs.vcs.VCS()
        self.items = []

        self.files_table = rabbitvcs.ui.widget.Table(
            self.get_widget("files_table"),
            [gobject.TYPE_BOOLEAN, rabbitvcs.ui.widget.TYPE_PATH, 
                gobject.TYPE_STRING, rabbitvcs.ui.widget.TYPE_STATUS,
                gobject.TYPE_STRING], 
            [rabbitvcs.ui.widget.TOGGLE_BUTTON, _("Path"), _("Extension"), 
                _("Text Status"), _("Property Status")],
            filters=[{
                "callback": rabbitvcs.ui.widget.path_filter,
                "user_data": {
                    "base_dir": base_dir,
                    "column": 1
                }
            }],
            callbacks={
                "row-activated":  self.on_files_table_row_activated,
                "mouse-event":   self.on_files_table_mouse_event,
                "key-event":     self.on_files_table_key_event,
                "row-toggled":   self.on_files_table_toggle_event
            },
            flags={
                "sortable": True, 
                "sort_on": 1
            }
        )
        self.files_table.allow_multiple()
        self.get_widget("toggle_show_unversioned").set_active(not self.SHOW_UNVERSIONED)
        self.message = rabbitvcs.ui.widget.TextView(
            self.get_widget("message"),
            (message and message or "")
        )

        self.paths = []
        for path in paths:
            if self.vcs.is_in_a_or_a_working_copy(path):
                self.paths.append(path)

    #
    # Helper functions
    # 

    def load(self):
        """
          - Gets a listing of file items that are valid for the commit window.
          - Determines which items should be "activated" by default
          - Populates the files table with the retrieved items
          - Updates the status area        
        """
        gtk.gdk.threads_enter()
        self.get_widget("status").set_text(_("Loading..."))
        gtk.gdk.threads_leave()

        self.items = self.vcs.get_items(self.paths, self.vcs.statuses_for_commit(self.paths))

        gtk.gdk.threads_enter()
        self.populate_files_table()
        gtk.gdk.threads_leave()

    # Overrides the GtkContextMenuCaller method
    def on_context_menu_command_finished(self):
        self.initialize_items()

    def should_item_be_activated(self, item):
        """
        Determines if a file should be activated or not
        """

        if (item.path in self.paths
                or item.is_versioned()
                and item.simple_content_status() != rabbitvcs.vcs.status.status_missing):
            return True

        return False

    def should_item_be_visible(self, item):
        show_unversioned = self.SHOW_UNVERSIONED
        
        if not show_unversioned:
            if not item.is_versioned():
               return False
       
        return True

    def initialize_items(self):
        """
        Initializes the activated cache and loads the file items in a new thread
        """
        
        try:
            thread.start_new_thread(self.load, ())
        except Exception, e:
            log.exception(e)

    def show_files_table_popup_menu(self, treeview, data):
        paths = self.files_table.get_selected_row_items(1)
        GtkFilesContextMenu(self, data, self.base_dir, paths).show()

    def delete_items(self, widget, data=None):
        paths = self.files_table.get_selected_row_items(1)
        if len(paths) > 0:
            proc = rabbitvcs.util.helper.launch_ui_window("delete", paths)
            self.rescan_after_process_exit(proc, paths)
            
    #
    # Event handlers
    #
    def on_refresh_clicked(self, widget):
        self.initialize_items()        

    def on_key_pressed(self, widget, data):
        if InterfaceView.on_key_pressed(self, widget, data):
            return True

        if (data.state & (gtk.gdk.CONTROL_MASK) and 
                gtk.gdk.keyval_name(data.keyval) == "Return"):
            self.on_ok_clicked(widget)
            return True
            
    def on_toggle_show_all_toggled(self, widget, data=None):
        self.TOGGLE_ALL = not self.TOGGLE_ALL
        self.changes.clear()
        for row in self.files_table.get_items():
            row[0] = self.TOGGLE_ALL
            self.changes[row[1]] = self.TOGGLE_ALL
            
    def on_toggle_show_unversioned_toggled(self, widget, data=None):

        self.SHOW_UNVERSIONED = not self.SHOW_UNVERSIONED
        
        self.populate_files_table()

        # Save this preference for future commits.
        if self.SETTINGS.get("general", "show_unversioned_files") != self.SHOW_UNVERSIONED:
            self.SETTINGS.set(
                "general", "show_unversioned_files",
                self.SHOW_UNVERSIONED
            )
            self.SETTINGS.write()

    def on_files_table_row_activated(self, treeview, event, col):
        paths = self.files_table.get_selected_row_items(1)
        pathrev1 = rabbitvcs.util.helper.create_path_revision_string(paths[0], "base")
        pathrev2 = rabbitvcs.util.helper.create_path_revision_string(paths[0], "working")
        proc = rabbitvcs.util.helper.launch_ui_window("diff", ["-s", pathrev1, pathrev2])
        self.rescan_after_process_exit(proc, paths)

    def on_files_table_key_event(self, treeview, data=None):
        if gtk.gdk.keyval_name(data.keyval) == "Delete":
            self.delete_items(treeview, data)

    def on_files_table_mouse_event(self, treeview, data=None):
        if data is not None and data.button == 3:
            self.show_files_table_popup_menu(treeview, data)

    def on_previous_messages_clicked(self, widget, data=None):
        dialog = rabbitvcs.ui.dialog.PreviousMessages()
        message = dialog.run()
        if message is not None:
            self.message.set_text(message)

    def populate_files_table(self):
        """
        First clears and then populates the files table based on the items
        retrieved in self.load()
        
        """
        
        self.files_table.clear()
        n = 0
        for item in self.items:
            if item.path in self.changes:
                checked = self.changes[item.path]
            else:
                checked = self.should_item_be_activated(item)

            if not self.should_item_be_visible(item):
                continue

            n += 1
            self.files_table.append([
                checked,
                item.path, 
                rabbitvcs.util.helper.get_file_extension(item.path),
                item.simple_content_status(),
                item.simple_metadata_status()
            ])
        self.get_widget("status").set_text(_("Found %d item(s)") % n)

class SVNCommit(Commit):
    def __init__(self, paths, base_dir=None, message=None):
        Commit.__init__(self, paths, base_dir, message)

        self.get_widget("commit_to_box").show()
        
        self.get_widget("to").set_text(
            self.vcs.svn().get_repo_url(self.base_dir)
        )

        self.items = None
        if len(self.paths):
            self.initialize_items()

    def on_ok_clicked(self, widget, data=None):
        items = self.files_table.get_activated_rows(1)
        self.hide()

        if len(items) == 0:
            self.close()
            return

        added = 0
        recurse = False
        for item in items:
            status = self.vcs.status(item, summarize=False).simple_content_status()
            try:
                if status == rabbitvcs.vcs.status.status_unversioned:
                    self.vcs.svn().add(item)
                    added += 1
                elif status == rabbitvcs.vcs.status.status_deleted:
                    recurse = True
                elif status == rabbitvcs.vcs.status.status_missing:
                    self.vcs.svn().update(item)
                    self.vcs.svn().remove(item)
            except Exception, e:
                log.exception(e)

        ticks = added + len(items)*2

        self.action = rabbitvcs.ui.action.SVNAction(
            self.vcs.svn(),
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.set_pbar_ticks(ticks)
        self.action.append(self.action.set_header, _("Commit"))
        self.action.append(self.action.set_status, _("Running Commit Command..."))
        self.action.append(
            rabbitvcs.util.helper.save_log_message, 
            self.message.get_text()
        ),
        self.action.append(self.do_commit, items, recurse)
        self.action.append(self.action.finish)
        self.action.start()

    def do_commit(self, items, recurse):
        # pysvn.Revision
        revision = self.vcs.svn().commit(items, self.message.get_text(), recurse=recurse)
        
        self.action.set_status(_("Completed Commit") + " at Revision: " + str(revision.number))

    def on_files_table_toggle_event(self, row, col):
        # Adds path: True/False to the dict
        self.changes[row[1]] = row[col]

class GitCommit(Commit):
    def __init__(self, paths, base_dir=None, message=None):
        Commit.__init__(self, paths, base_dir, message)

        self.git = self.vcs.git(paths[0])

        self.items = None
        if len(self.paths):
            self.initialize_items()

    def on_ok_clicked(self, widget, data=None):
        items = self.files_table.get_activated_rows(1)
        self.hide()

        if len(items) == 0:
            self.close()
            return

        staged = 0
        for item in items:
            try:
                status = self.vcs.status(item, summarize=False).simple_content_status()
                if status == rabbitvcs.vcs.status.status_missing:
                    self.git.checkout([item])
                    self.git.remove(item)
                else:
                    self.git.stage(item)
                    staged += 1
            except Exception, e:
                log.exception(e)

        ticks = staged + len(items)*2

        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.set_pbar_ticks(ticks)
        self.action.append(self.action.set_header, _("Commit"))
        self.action.append(self.action.set_status, _("Running Commit Command..."))
        self.action.append(
            rabbitvcs.util.helper.save_log_message,
            self.message.get_text()
        )
        self.action.append(
            self.git.commit, 
            self.message.get_text()
        )
        self.action.append(self.action.set_status, _("Completed Commit"))
        self.action.append(self.action.finish)
        self.action.start() 

    def on_files_table_toggle_event(self, row, col):
        # Adds path: True/False to the dict
        self.changes[row[1]] = row[col]


class MercurialCommit(Commit):
    def __init__(self, paths, base_dir=None, message=None):
        Commit.__init__(self, paths, base_dir, message)

        self.mercurial = self.vcs.mercurial(paths[0])

        options_box = self.get_widget("options_box")
        
        self.close_branch = gtk.CheckButton(_("Close Branch"))
        options_box.pack_start(self.close_branch, False, False, 0)
        self.close_branch.show()

        self.items = None
        if len(self.paths):
            self.initialize_items()

    def on_ok_clicked(self, widget, data=None):
        items = self.files_table.get_activated_rows(1)
        self.hide()

        if len(items) == 0:
            self.close()
            return

        staged = 0
        for item in items:
            try:
                status = self.vcs.status(item, summarize=False).simple_content_status()
                if status == rabbitvcs.vcs.status.status_missing:
                    self.mercurial.revert([item])
                    self.mercurial.remove(item)
                else:
                    self.mercurial.add(item)
                    staged += 1
            except Exception, e:
                log.exception(e)

        ticks = staged + len(items)*2

        self.action = rabbitvcs.ui.action.MercurialAction(
            self.mercurial,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.set_pbar_ticks(ticks)
        self.action.append(self.action.set_header, _("Commit"))
        self.action.append(self.action.set_status, _("Running Commit Command..."))
        self.action.append(
            rabbitvcs.util.helper.save_log_message, 
            self.message.get_text()
        )
        self.action.append(
            self.mercurial.commit, 
            self.message.get_text()
        )
        self.action.append(self.action.set_status, _("Completed Commit"))
        self.action.append(self.action.finish)
        self.action.start()

    def on_files_table_toggle_event(self, row, col):
        # Adds path: True/False to the dict
        self.changes[row[1]] = row[col]

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNCommit, 
    rabbitvcs.vcs.VCS_GIT: GitCommit, 
    rabbitvcs.vcs.VCS_MERCURIAL: MercurialCommit
}

def commit_factory(paths, base_dir=None, message=None):
    guess = rabbitvcs.vcs.guess(paths[0])
    return classes_map[guess["vcs"]](paths, base_dir, message)

if __name__ == "__main__":
    from rabbitvcs.ui import main, BASEDIR_OPT
    (options, paths) = main(
        [BASEDIR_OPT, (["-m", "--message"], {"help":"add a commit log message"})],
        usage="Usage: rabbitvcs commit [path1] [path2] ..."
    )

    window = commit_factory(paths, options.base_dir, message=options.message)
    window.register_gtk_quit()
    gtk.main()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceNonView
from rabbitvcs.ui.action import SVNAction
import rabbitvcs.vcs
from rabbitvcs.util.log import Log

log = Log("rabbitvcs.ui.delete")

from rabbitvcs import gettext
_ = gettext.gettext

class Delete(InterfaceNonView):
    """
    This class provides a handler to Delete functionality.
    
    """

    def __init__(self, paths):
        InterfaceNonView.__init__(self)
        self.paths = paths
        self.vcs = rabbitvcs.vcs.VCS()

    def start(self):
    
        # From the given paths, determine which are versioned and which are not
        versioned = []
        unversioned = []
        for path in self.paths:
            if self.vcs.is_versioned(path):
                versioned.append(path)
            elif os.path.exists(path):
                unversioned.append(path)
        
        # If there are unversioned files, confirm that the user wants to
        # delete those.  Default to true.
        result = True
        if unversioned:
            item = None
            if len(unversioned) == 1:
                item = unversioned[0]
            confirm = rabbitvcs.ui.dialog.DeleteConfirmation(item)
            result = confirm.run()

        # If the user wants to continue (or there are no unversioned files)
        # remove or delete the given files
        if result == gtk.RESPONSE_OK or result == True:
            if versioned:
                try:
                    self.vcs_remove(versioned, force=True)
                except Exception, e:
                    log.exception()
                    return
            
            if unversioned:
                for path in unversioned:
                    rabbitvcs.util.helper.delete_item(path)

class SVNDelete(Delete):
    def __init__(self, paths):
        Delete.__init__(self, paths)
    
    def vcs_remove(self, paths, **kwargs):
        if rabbitvcs.vcs.guess(paths[0])["vcs"] == rabbitvcs.vcs.VCS_SVN:
            self.vcs.svn().remove(paths, **kwargs)

class GitDelete(Delete):
    def __init__(self, paths):
        Delete.__init__(self, paths)
    
    def vcs_remove(self, paths, **kwargs):
        if rabbitvcs.vcs.guess(paths[0])["vcs"] == rabbitvcs.vcs.VCS_GIT:
            self.vcs.git(paths[0]).remove(paths)

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNDelete,
    rabbitvcs.vcs.VCS_GIT: GitDelete
}

def delete_factory(paths):
    guess = rabbitvcs.vcs.guess(paths[0])
    return classes_map[guess["vcs"]](paths)

if __name__ == "__main__":
    from rabbitvcs.ui import main
    (options, paths) = main(usage="Usage: rabbitvcs delete [path1] [path2] ...")

    window = delete_factory(paths)
    window.register_gtk_quit()
    window.start()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os

import pygtk
import gobject
import gtk
import pango

from datetime import datetime
import time

from rabbitvcs.ui import InterfaceView
from rabbitvcs.ui.action import GitAction
import rabbitvcs.ui.widget
import rabbitvcs.util.helper
import rabbitvcs.vcs

from rabbitvcs import gettext
_ = gettext.gettext

class GitClean(InterfaceView):
    """
    Provides a UI to clean your repository of untracked files
    
    """
    
    def __init__(self, path):
        InterfaceView.__init__(self, "clean", "Clean")
        self.vcs = rabbitvcs.vcs.VCS()
        self.git = self.vcs.git(path)
        self.path = path

    def on_ok_clicked(self, widget):
        remove_dir = self.get_widget("remove_directories").get_active()
        remove_ignored_too = self.get_widget("remove_ignored_too").get_active()
        remove_only_ignored = self.get_widget("remove_only_ignored").get_active()
        dry_run = self.get_widget("dryrun").get_active()
        force = self.get_widget("force").get_active()

        self.hide()
        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.append(self.action.set_header, _("Clean"))
        self.action.append(self.action.set_status, _("Running Clean Command..."))
        self.action.append(
            self.git.clean,
            self.path,
            remove_dir,
            remove_ignored_too,
            remove_only_ignored,
            dry_run,
            force
        )
        self.action.append(self.action.set_status, _("Completed Clean"))
        self.action.append(self.action.finish)
        self.action.start()

    def on_remove_ignored_too_toggled(self, widget):
        remove_ignored_too = self.get_widget("remove_ignored_too")
        remove_only_ignored = self.get_widget("remove_only_ignored")
        
        if remove_ignored_too.get_active():
            remove_only_ignored.set_active(False)

    def on_remove_only_ignored_toggled(self, widget):
        remove_ignored_too = self.get_widget("remove_ignored_too")
        remove_only_ignored = self.get_widget("remove_only_ignored")
        
        if remove_only_ignored.get_active():
            remove_ignored_too.set_active(False)

if __name__ == "__main__":
    from rabbitvcs.ui import main
    (options, paths) = main(usage="Usage: rabbitvcs clean path")
    
    window = GitClean(paths[0])
    window.register_gtk_quit()
    gtk.main()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import thread

import pygtk
import gobject
import gtk
import os
from shutil import rmtree
import tempfile

from rabbitvcs import TEMP_DIR_PREFIX
from rabbitvcs.ui import InterfaceNonView
import rabbitvcs.vcs
from rabbitvcs.ui.action import SVNAction, GitAction
import rabbitvcs.util.helper
from rabbitvcs.util.log import Log

log = Log("rabbitvcs.ui.diff")

from rabbitvcs import gettext
_ = gettext.gettext

gtk.gdk.threads_init()

class Diff(InterfaceNonView):
    def __init__(self, path1, revision1=None, path2=None, revision2=None, 
            sidebyside=False):
        InterfaceNonView.__init__(self)
        
        self.vcs = rabbitvcs.vcs.VCS()

        self.path1 = path1
        self.path2 = path2
        self.sidebyside = sidebyside

        self.temp_dir = tempfile.mkdtemp(prefix=TEMP_DIR_PREFIX)

        if path2 is None:
            self.path2 = path1
        
        self.dialog = None
        
    def launch(self):
        
        if self.sidebyside:
            self.launch_sidebyside_diff()
        else:
            self.launch_unified_diff()
        
        self.stop_loading()
        
    def _build_export_path(self, index, revision, path):
        dest = rabbitvcs.util.helper.get_tmp_path("rabbitvcs-%s-%s-%s" % (str(index), str(revision)[:5], os.path.basename(path)))
        if os.path.exists(dest):
            if os.path.isdir(dest):
                rmtree(dest, ignore_errors=True)
            else:
                os.remove(dest)

        return dest

    def start_loading(self):
        self.dialog = rabbitvcs.ui.dialog.Loading()
        self.dialog.run()
    
    def stop_loading(self):
        
        # Sometimes the launching will be too fast, and the dialog we're trusted with
        # cleaning up, may not even have been created!
        while self.dialog == None:
            # Wait for dialog's creation.
            pass
        
        self.dialog.close()
        self.dialog = None

class SVNDiff(Diff):
    def __init__(self, path1, revision1=None, path2=None, revision2=None,
            sidebyside=False):
        Diff.__init__(self, path1, revision1, path2, revision2, sidebyside)

        self.svn = self.vcs.svn()

        self.revision1 = self.get_revision_object(revision1, "base")
        self.revision2 = self.get_revision_object(revision2, "working")

        try:
            thread.start_new_thread(self.launch, ())
        except Exception, e:
            log.exception(e)
        
        self.start_loading()

    def get_revision_object(self, value, default):
        # If value is a rabbitvcs Revision object, return it
        if hasattr(value, "is_revision_object"):
            return value
        
        # If value is None, use the default
        if value is None:
            return self.svn.revision(default)          

        # If the value is an integer number, return a numerical revision object
        # otherwise, a string revision value has been passed, use that as "kind"
        try:
            value = int(value)
            return self.svn.revision("number", value)
        except ValueError:
            # triggered when passed a string
            return self.svn.revision(value)

    def launch_unified_diff(self):
        """
        Launch diff as a unified diff in a text editor or .diff viewer
        
        """
        
        action = SVNAction(
            self.svn,
            notification=False,
            run_in_thread=False
        )
        
        diff_text = action.run_single(
            self.svn.diff,
            self.temp_dir,
            self.path1,
            self.revision1,
            self.path2,
            self.revision2
        )
        if diff_text is None:
            diff_text = ""

        fh = tempfile.mkstemp("-rabbitvcs-" + str(self.revision1) + "-" + str(self.revision2) + ".diff")
        os.write(fh[0], diff_text)
        os.close(fh[0])
        rabbitvcs.util.helper.open_item(fh[1])
        
    def launch_sidebyside_diff(self):
        """
        Launch diff as a side-by-side comparison using our comparison tool
        
        """

        action = SVNAction(
            self.svn,
            notification=False,
            run_in_thread=False
        )

        if self.revision1.kind == "working":
            dest1 = self.path1
        else:
            dest1 = self._build_export_path(1, self.revision1, self.path1)
            action.run_single(
                self.svn.export, 
                self.path1, 
                dest1, 
                self.revision1
            )
            action.stop_loader()
    
        if self.revision2.kind == "working":
            dest2 = self.path2
        else:
            dest2 = self._build_export_path(2, self.revision2, self.path2)
            action.run_single(
                self.svn.export, 
                self.path2, 
                dest2, 
                self.revision2
            )
            action.stop_loader()
    
        rabbitvcs.util.helper.launch_diff_tool(dest1, dest2)

class GitDiff(Diff):
    def __init__(self, path1, revision1=None, path2=None, revision2=None,
            sidebyside=False):
        Diff.__init__(self, path1, revision1, path2, revision2, sidebyside)

        self.git = self.vcs.git(path1)

        self.revision1 = self.get_revision_object(revision1, "HEAD")
        self.revision2 = self.get_revision_object(revision2, "WORKING")

        try:
            thread.start_new_thread(self.launch, ())
        except Exception, e:
            log.exception(e)
        
        self.start_loading()

    def get_revision_object(self, value, default):
        # If value is a rabbitvcs Revision object, return it
        if hasattr(value, "is_revision_object"):
            return value        

        value_to_pass = value
        if not value_to_pass:
            value_to_pass = default

        # triggered when passed a string
        return self.git.revision(value_to_pass)

    def save_diff_to_file(self, path, data):
        dirname = os.path.dirname(path)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        
        if not data:
            data = ""
        
        file = open(path, "wb")
        try:
            file.write(data)
        except Exception, e:
            log.exception(e)
        finally:
            file.close()

    def launch_unified_diff(self):
        """
        Launch diff as a unified diff in a text editor or .diff viewer
        
        """
        
        action = GitAction(
            self.git,
            notification=False,
            run_in_thread=False
        )

        diff_text = action.run_single(
            self.git.diff,
            self.path1,
            self.revision1,
            self.path2,
            self.revision2
        )
        if diff_text is None:
            diff_text = ""

        fh = tempfile.mkstemp("-rabbitvcs-" + str(self.revision1)[:5] + "-" + str(self.revision2)[:5] + ".diff")
        os.write(fh[0], diff_text)
        os.close(fh[0])
        rabbitvcs.util.helper.open_item(fh[1])

    def launch_sidebyside_diff(self):
        """
        Launch diff as a side-by-side comparison using our comparison tool
        
        """
        
        action = GitAction(
            self.git,
            notification=False,
            run_in_thread=False
        )
        
        if self.revision1.kind != "WORKING":
            dest1 = self._build_export_path(1, self.revision1, self.path1)
            self.save_diff_to_file(dest1, action.run_single(
                self.git.show, 
                self.path1,
                self.revision1
            ))
        else:
            dest1 = self.path1

        if self.revision2.kind != "WORKING":
            dest2 = self._build_export_path(2, self.revision2, self.path2)
            self.save_diff_to_file(dest2, action.run_single(
                self.git.show, 
                self.path2,
                self.revision2
            ))
        else:
            dest2 = self.path2

        rabbitvcs.util.helper.launch_diff_tool(dest1, dest2)

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNDiff,
    rabbitvcs.vcs.VCS_GIT: GitDiff
}

def diff_factory(vcs, path1, revision_obj1, path2=None, revision_obj2=None, sidebyside=False):
    if not vcs:
        guess = rabbitvcs.vcs.guess(path1)
        vcs = guess["vcs"]

    return classes_map[vcs](path1, revision_obj1, path2, revision_obj2, sidebyside)

if __name__ == "__main__":
    from rabbitvcs.ui import main, VCS_OPT
    (options, args) = main([
        (["-s", "--sidebyside"], {
            "help":     _("View diff as side-by-side comparison"), 
            "action":   "store_true", 
            "default":  False
        }), VCS_OPT],
        usage="Usage: rabbitvcs diff [url1@rev1] [url2@rev2]"
    )
    
    pathrev1 = rabbitvcs.util.helper.parse_path_revision_string(args.pop(0))
    pathrev2 = (None, None)
    if len(args) > 0:
        pathrev2 = rabbitvcs.util.helper.parse_path_revision_string(args.pop(0))

    diff_factory(options.vcs, pathrev1[0], pathrev1[1], pathrev2[0], pathrev2[1], sidebyside=options.sidebyside)



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView
import rabbitvcs.ui.widget
import rabbitvcs.util.helper
from rabbitvcs.util.contextmenu import GtkContextMenu
from rabbitvcs.util.contextmenuitems import *
import rabbitvcs.ui.action
from rabbitvcs.ui.dialog import MessageBox
from rabbitvcs import gettext
_ = gettext.gettext

class Changes(InterfaceView):
    """
    Show how files and folders are different between revisions.
    
        TODO:
            - Deal with the revision arguments in a smarter way so we can pass
                in revisions like HEAD.  Currently, if a revision is passed it
                assumes it is a number
    """
    selected_rows = []
    MORE_ACTIONS_ITEMS = [
        _("More Actions..."),
        _("View unified diff")
    ]

    def __init__(self, path1=None, revision1=None, path2=None, revision2=None):
        InterfaceView.__init__(self, "changes", "Changes")
        
        self.vcs = rabbitvcs.vcs.VCS()

        self.MORE_ACTIONS_CALLBACKS = [
            None,
            self.on_more_actions_view_unified_diff
        ]

        self.more_actions = rabbitvcs.ui.widget.ComboBox(
            self.get_widget("more_actions"),
            self.MORE_ACTIONS_ITEMS
        )
        self.more_actions.set_active(0)

        repo_paths = rabbitvcs.util.helper.get_repository_paths()
        self.first_urls = rabbitvcs.ui.widget.ComboBox(
            self.get_widget("first_urls"), 
            repo_paths
        )
        self.first_urls_browse = self.get_widget("first_urls_browse")

        self.second_urls = rabbitvcs.ui.widget.ComboBox(
            self.get_widget("second_urls"), 
            repo_paths
        )
        self.second_urls_browse = self.get_widget("second_urls_browse")
        
    #
    # UI Signal Callback Methods
    #

    def on_close_clicked(self, widget):
        self.close()

    def on_refresh_clicked(self, widget):
        self.load()
    
    def on_first_urls_changed(self, widget, data=None):
        self.check_first_urls()
        self.check_refresh_button()

    def on_second_urls_changed(self, widget, data=None):
        self.check_second_urls()
        self.check_refresh_button()

    def on_first_urls_browse_clicked(self, widget, data=None):
        pass

    def on_first_repo_chooser_closed(self, new_url):
        self.first_urls.set_child_text(new_url)
        self.check_first_urls()
        self.check_refresh_button()

    def on_second_urls_browse_clicked(self, widget, data=None):
        pass

    def on_second_repo_chooser_closed(self, new_url):
        self.second_urls.set_child_text(new_url)
        self.check_second_urls()
        self.check_refresh_button()

    def on_changes_table_cursor_changed(self, treeview, data=None):
        self.on_changes_table_event(treeview, data)

    def on_changes_table_button_released(self, treeview, data=None):
        self.on_changes_table_event(treeview, data)

    def on_changes_table_event(self, treeview, data=None):
        selection = treeview.get_selection()
        (liststore, indexes) = selection.get_selected_rows()

        self.selected_rows = []
        for tup in indexes:
            self.selected_rows.append(tup[0])

        if data is not None and data.button == 3:
            self.show_changes_table_popup_menu(treeview, data)

    def on_more_actions_changed(self, widget, data=None):
        index = self.more_actions.get_active()
        if index < 0:
            return
            
        callback = self.MORE_ACTIONS_CALLBACKS[index]
        
        if callback is not None:
            callback()

    def on_changes_table_row_doubleclicked(self, treeview, data=None, col=None):
        selection = treeview.get_selection()
        (liststore, indexes) = selection.get_selected_rows()

        self.selected_rows = []
        for tup in indexes:
            self.selected_rows.append(tup[0])

        self.view_selected_diff(sidebyside=True)

    #
    # Helper methods
    #
    
    def get_first_revision(self):
        return self.first_revision_selector.get_revision_object()
    
    def get_second_revision(self):
        return self.second_revision_selector.get_revision_object()

    def show_changes_table_popup_menu(self, treeview, data):
        ChangesContextMenu(self, data).show()
    
    def check_ui(self):
        self.check_first_urls()
        self.check_second_urls()
        self.check_refresh_button()
    
    def can_first_browse_urls(self):
        return (self.first_urls.get_active_text() != "")
    
    def can_second_browse_urls(self):
        return (self.second_urls.get_active_text() != "")
    
    def check_refresh_button(self):
        can_click_refresh = (
            self.can_first_browse_urls()
            and self.can_second_browse_urls()
        )
        
        self.get_widget("refresh").set_sensitive(can_click_refresh)
    
    def check_first_urls(self):
        can_browse_urls = self.can_first_browse_urls()
        self.first_urls_browse.set_sensitive(can_browse_urls)
        
    def check_second_urls(self):
        can_browse_urls = self.can_second_browse_urls()
        self.second_urls_browse.set_sensitive(can_browse_urls)


    def enable_more_actions(self):
        self.more_actions.set_sensitive(True)

    def disable_more_actions(self):
        self.more_actions.set_sensitive(False)
    
    def view_selected_diff(self, sidebyside=False):
        for row in self.selected_rows:
            url1 = self.changes_table.get_row(row)[0]
            url2 = url1
            if url1 == ".":
                url1 = ""
                url2 = ""

            url1 = rabbitvcs.util.helper.url_join(self.first_urls.get_active_text(), url1)
            url2 = rabbitvcs.util.helper.url_join(self.second_urls.get_active_text(), url2)
            rev1 = self.get_first_revision()
            rev2 = self.get_second_revision()
            
            rabbitvcs.util.helper.launch_ui_window("diff", [
                "%s@%s" % (url1, unicode(rev1)),
                "%s@%s" % (url2, unicode(rev2)),
                "%s" % (sidebyside and "-s" or ""),
                "--vcs=%s" % self.get_vcs_name()
            ])
        
        
    #
    # More Actions callbacks
    #

    def on_more_actions_view_unified_diff(self):
        url1 = self.first_urls.get_active_text()
        rev1 = self.get_first_revision()
        rev2 = self.get_second_revision()        
        url2 = self.second_urls.get_active_text()

        rabbitvcs.util.helper.launch_ui_window("diff", [
            "%s@%s" % (url1, unicode(rev1)),
            "%s@%s" % (url2, unicode(rev2)),
            "--vcs=%s" % self.get_vcs_name()
        ])

    def get_vcs_name(self):
        vcs = rabbitvcs.vcs.VCS_DUMMY
        if hasattr(self, "svn"):
            vcs = rabbitvcs.vcs.VCS_SVN
        elif hasattr(self, "git"):
            vcs = rabbitvcs.vcs.VCS_GIT

        return vcs

class SVNChanges(Changes):
    def __init__(self, path1=None, revision1=None, path2=None, revision2=None):
        Changes.__init__(self, path1, revision1, path2, revision2)

        self.svn = self.vcs.svn()

        if path1 is not None:
            self.first_urls.set_child_text(self.svn.get_repo_url(path1))
            
        if path2 is not None:
            self.second_urls.set_child_text(self.svn.get_repo_url(path2))
        elif path1 is not None:
            self.second_urls.set_child_text(self.svn.get_repo_url(path1))

        self.first_revision_selector = rabbitvcs.ui.widget.RevisionSelector(
            self.get_widget("first_revision_container"),
            self.svn,
            revision=revision1,
            url_combobox=self.first_urls
        )

        self.second_revision_selector = rabbitvcs.ui.widget.RevisionSelector(
            self.get_widget("second_revision_container"),
            self.svn,
            revision=revision2,
            url_combobox=self.second_urls
        )

        self.changes_table = rabbitvcs.ui.widget.Table(
            self.get_widget("changes_table"),
            [gobject.TYPE_STRING, gobject.TYPE_STRING, 
                gobject.TYPE_STRING], 
            [_("Path"), _("Change"), _("Property Change")],
            flags={
                "sortable": True,
                "sort_on": 1
            },
            callbacks={
                "mouse-event":   self.on_changes_table_button_released
            }
        )

        self.check_ui()
        
        if path1 and revision1 and path2 and revision2:
            self.load()

    def load(self):
        first_url = self.first_urls.get_active_text()
        first_rev = self.get_first_revision()
        second_rev = self.get_second_revision()        
        second_url = self.second_urls.get_active_text()

        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            notification=False
        )
        self.action.append(self.disable_more_actions)
        self.action.append(
            self.svn.diff_summarize,
            first_url,
            first_rev,
            second_url,
            second_rev
        )
        self.action.append(rabbitvcs.util.helper.save_repository_path, first_url)
        self.action.append(rabbitvcs.util.helper.save_repository_path, second_url)
        self.action.append(self.populate_table)
        self.action.append(self.enable_more_actions)
        self.action.start()

    def populate_table(self):
        # returns a list of dicts(path, summarize_kind, node_kind, prop_changed)
        summary = self.action.get_result(1)

        self.changes_table.clear()
        for item in summary:
            prop_changed = (item["prop_changed"] == 1 and _("Yes") or _("No"))
            
            path = item["path"]
            if path == "":
                path = "."
                
            self.changes_table.append([
                path,
                item["summarize_kind"],
                prop_changed
            ])

    def on_first_urls_browse_clicked(self, widget, data=None):
        from rabbitvcs.ui.browser import SVNBrowserDialog
        SVNBrowserDialog(self.first_urls.get_active_text(), 
            callback=self.on_first_repo_chooser_closed)

    def on_second_urls_browse_clicked(self, widget, data=None):
        from rabbitvcs.ui.browser import SVNBrowserDialog
        SVNBrowserDialog(self.second_urls.get_active_text(), 
            callback=self.on_second_repo_chooser_closed)

class GitChanges(Changes):
    def __init__(self, path1=None, revision1=None, path2=None, revision2=None):
        Changes.__init__(self, path1, revision1, path2, revision2)

        self.git = self.vcs.git(path1)
        
        self.first_urls_browse.hide()
        self.second_urls_browse.hide()

        if path1 is not None:
            self.first_urls.set_child_text(path1)
            
        if path2 is not None:
            self.second_urls.set_child_text(path2)
        elif path1 is not None:
            self.second_urls.set_child_text(path1)

        self.first_revision_selector = rabbitvcs.ui.widget.RevisionSelector(
            self.get_widget("first_revision_container"),
            self.git,
            revision=revision1,
            url_combobox=self.first_urls
        )

        self.second_revision_selector = rabbitvcs.ui.widget.RevisionSelector(
            self.get_widget("second_revision_container"),
            self.git,
            revision=revision2,
            url_combobox=self.second_urls
        )

        self.changes_table = rabbitvcs.ui.widget.Table(
            self.get_widget("changes_table"),
            [gobject.TYPE_STRING, gobject.TYPE_STRING], 
            [_("Path"), _("Change")]
        )

        self.check_ui()
        
        if path1 and revision1 and path2 and revision2:
            self.load()

    def load(self):
        first_url = self.first_urls.get_active_text()
        first_rev = self.get_first_revision()
        second_rev = self.get_second_revision()        
        second_url = self.second_urls.get_active_text()

        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            notification=False
        )
        self.action.append(self.disable_more_actions)
        self.action.append(
            self.git.diff_summarize,
            first_url,
            first_rev,
            second_url,
            second_rev
        )
        self.action.append(rabbitvcs.util.helper.save_repository_path, first_url)
        self.action.append(rabbitvcs.util.helper.save_repository_path, second_url)
        self.action.append(self.populate_table)
        self.action.append(self.enable_more_actions)
        self.action.start()

    def populate_table(self):
        # returns a list of dicts(path, summarize_kind, node_kind, prop_changed)
        summary = self.action.get_result(1)

        self.changes_table.clear()
        for item in summary:
            self.changes_table.append([
                item.path,
                item.action
            ])


class MenuOpenFirst(MenuItem):
    identifier = "RabbitVCS::Open_First"
    label = _("Open from first revision")
    icon = gtk.STOCK_OPEN

class MenuOpenSecond(MenuItem):
    identifier = "RabbitVCS::Open_Second"
    label = _("Open from second revision") 
    icon = gtk.STOCK_OPEN
    
class MenuViewDiff(MenuItem):
    identifier = "RabbitVCS::View_Diff"
    label = _("View unified diff(s)")   
    icon = "rabbitvcs-diff"

class MenuCompare(MenuItem):
    identifier = "RabbitVCS::Compare"
    label = _("Compare side by side")
    icon = "rabbitvcs-compare"

class ChangesContextMenuConditions:
    def __init__(self, caller, vcs):
        self.caller = caller
        self.vcs = vcs

    def open_first(self):
        return (
            len(self.caller.selected_rows) == 1
        )
    
    def open_second(self):
        return (
            len(self.caller.selected_rows) == 1 
            and (
                str(self.caller.get_first_revision()) != str(self.caller.get_second_revision())
                or self.caller.first_urls.get_active_text() != self.caller.second_urls.get_active_text()
            )
        )

    def view_diff(self):
        return (
            len(self.caller.selected_rows) > 0
        )

    def compare(self):
        return (
            len(self.caller.selected_rows) > 0
        )

class ChangesContextMenuCallbacks:
    def __init__(self, caller, vcs):
        self.caller = caller
        self.vcs = vcs

    def open_first(self, widget, data=None):
        path = self.caller.changes_table.get_row(self.caller.selected_rows[0])[0]
        if path == ".":
            path = ""

        url = rabbitvcs.util.helper.url_join(self.caller.first_urls.get_active_text(), path)
        rev = self.caller.get_first_revision()

        rabbitvcs.util.helper.launch_ui_window("open", [
            "--vcs=%s" % self.caller.get_vcs_name(),
            url,
            "-r%s" % unicode(rev)            
        ])

    def open_second(self, widget, data=None):
        path = self.caller.changes_table.get_row(self.caller.selected_rows[0])[0]
        if path == ".":
            path = ""

        url = rabbitvcs.util.helper.url_join(self.caller.second_urls.get_active_text(), path)
        rev = self.caller.get_second_revision()
        rabbitvcs.util.helper.launch_ui_window("open", [
            "--vcs=%s" % self.caller.get_vcs_name(),
            url,
            "-r%s" % unicode(rev)            
        ])

    def view_diff(self, widget, data=None):
        self.caller.view_selected_diff()

    def compare(self, widget, data=None):
        for row in self.caller.selected_rows:
            url1 = self.caller.changes_table.get_row(row)[0]
            url2 = url1
            if url1 == ".":
                url1 = ""
                url2 = ""

            url1 = rabbitvcs.util.helper.url_join(self.caller.first_urls.get_active_text(), url1)
            url2 = rabbitvcs.util.helper.url_join(self.caller.second_urls.get_active_text(), url2)
            rev1 = self.caller.get_first_revision()
            rev2 = self.caller.get_second_revision()
            
            rabbitvcs.util.helper.launch_ui_window("diff", [
                "%s@%s" % (url1, unicode(rev1)), 
                "%s@%s" % (url2, unicode(rev2)), 
                "-s",
                "--vcs=%s" % self.caller.get_vcs_name()
            ])

class ChangesContextMenu:
    """
    Defines context menu items for a table with files
    
    """
    def __init__(self, caller, event):
        """    
        @param  caller: The calling object
        @type   caller: object
        
        """        
        self.caller = caller
        self.event = event
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()

        self.conditions = ChangesContextMenuConditions(
            self.caller,
            self.vcs
        )
        
        self.callbacks = ChangesContextMenuCallbacks(
            self.caller,
            self.vcs
        )

        # The first element of each tuple is a key that matches a
        # ContextMenuItems item.  The second element is either None when there
        # is no submenu, or a recursive list of tuples for desired submenus.
        self.structure = [
            (MenuOpenFirst, None),
            (MenuOpenSecond, None),
            (MenuViewDiff, None),
            (MenuCompare, None)
        ]
        
    def show(self):
        if len(self.caller.selected_rows) == 0:
            return
            
        context_menu = GtkContextMenu(self.structure, self.conditions, self.callbacks)
        context_menu.show(self.event)

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNChanges,
    rabbitvcs.vcs.VCS_GIT: GitChanges
}

def changes_factory(vcs, path1=None, revision1=None, path2=None, revision2=None):
    if not vcs:
        guess = rabbitvcs.vcs.guess(path1)
        vcs = guess["vcs"]
        
    return classes_map[vcs](path1, revision1, path2, revision2)

if __name__ == "__main__":
    from rabbitvcs.ui import main, VCS_OPT
    (options, args) = main(
        [VCS_OPT],
        usage="Usage: rabbitvcs changes [url1@rev1] [url2@rev2]"
    )
    
    pathrev1 = rabbitvcs.util.helper.parse_path_revision_string(args.pop(0))
    pathrev2 = (None, None)
    if len(args) > 0:
        pathrev2 = rabbitvcs.util.helper.parse_path_revision_string(args.pop(0))

    window = changes_factory(options.vcs, pathrev1[0], pathrev1[1], pathrev2[0], pathrev2[1])
    window.register_gtk_quit()
    gtk.main()



#
# test/stage.py
#

import os
from shutil import rmtree
from sys import argv
from optparse import OptionParser

from gittyup.client import GittyupClient
from gittyup.objects import *
from util import touch, change
from gittyup.config import GittyupLocalFallbackConfig, GittyupConfig, GittyupSystemConfig

parser = OptionParser()
parser.add_option("-c", "--cleanup", action="store_true", default=False)
(options, args) = parser.parse_args(argv)

DIR = "config"

if options.cleanup:
    rmtree(DIR, ignore_errors=True)

    print "config.py clean"
else:
    if os.path.isdir(DIR):
        raise SystemExit("This test script has already been run.  Please call this script with --cleanup to start again")

    g = GittyupClient(DIR, create=True)
    c = GittyupLocalFallbackConfig(DIR)
    
    # Create config items
    c.set("core", "filemode", True)
    c.set("core", "repositoryformatversion", 0)

    # Add comments
    c.set_comment("core", "filemode", ["Regular comment"])
    c.set_inline_comment("core", "repositoryformatversion", "inline repo format comment")

    # Create new section/items and then rename them
    c.set("newsection", "newitem", "Val A")
    c.rename_section("newsection", "newsection_RE")    
    c.rename("newsection_RE", "newitem", "newitem_RE")
    c.write()

    del c
    
    c = GittyupLocalFallbackConfig(DIR)

    assert (c.has("newsection_RE", "newitem_RE"))
    assert (c.get_comment("core", "filemode")[0].find("Regular comment") != -1)
    assert (c.get_inline_comment("core", "repositoryformatversion").find("inline repo format comment") != -1)
    
    del c

    c = GittyupConfig("./data/config/config.example")
    
    assert (c.has("diff", "renames"))
    
    del c
    
    c = GittyupSystemConfig()
    c.set("section", "key", "value")
    c.write()

    print "config.py pass"



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import thread
import threading

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView
from rabbitvcs.util.contextmenu import GtkFilesContextMenu, \
    GtkContextMenuCaller, GtkFilesContextMenuConditions, GtkContextMenu
from rabbitvcs.util.contextmenuitems import MenuItem, MenuUpdate, \
    MenuSeparator
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.ui.action
import rabbitvcs.util.helper
from rabbitvcs.util.log import Log
from rabbitvcs.util.decorators import gtk_unsafe

log = Log("rabbitvcs.ui.checkmods")

from rabbitvcs import gettext
_ = gettext.gettext

gtk.gdk.threads_init()

class SVNCheckForModifications(InterfaceView):
    """
    Provides a way for the user to see what files have been changed on the 
    repository.
    
    """
    
    def __init__(self, paths, base_dir=None):
        InterfaceView.__init__(self, "checkmods", "CheckMods")

        self.paths = paths
        self.base_dir = base_dir
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()
        self.notebook = self.get_widget("notebook")

        self.local_mods = SVNCheckLocalModifications(self, \
                                                         self.vcs, \
                                                         self.paths, \
                                                         self.base_dir)
        self.remote_mods = SVNCheckRemoteModifications(self, \
                                                           self.vcs, \
                                                           self.paths, \
                                                           self.base_dir)

        self.remote_refreshed = False

        self.load()

    def on_refresh_clicked(self, widget):
        if self.notebook.get_current_page() == 0:
            self.local_mods.refresh()
        else:
            self.remote_mods.refresh()

    def on_notebook_switch_page(self, page, data, page_num):
        if page_num == 1 and self.remote_refreshed == False:
            self.remote_mods.refresh()
            self.remote_refreshed = True

    #
    # Helper methods
    #
    
    def load(self):
        self.local_mods.refresh()

class SVNCheckLocalModifications(GtkContextMenuCaller):
    def __init__(self, caller, vcs, paths, base_dir):
        self.caller = caller
        self.vcs = vcs
        self.svn = vcs.svn()
        self.items = None
        self.paths = paths
        self.base_dir = base_dir

        self.files_table = rabbitvcs.ui.widget.Table(
            self.caller.get_widget("local_files_table"), 
            [gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING], 
            [_("Path"), _("Status"), _("Extension")],
            filters=[{
                "callback": rabbitvcs.ui.widget.path_filter,
                "user_data": {
                    "base_dir": base_dir,
                    "column": 0
                }
            }],
            callbacks={
                "row-activated":  self.on_files_table_row_activated,
                "mouse-event":   self.on_files_table_mouse_event
            }
        )

    def on_files_table_row_activated(self, treeview, event, col):
        paths = self.files_table.get_selected_row_items(0)
        self.diff_local(paths[0])

    def on_files_table_mouse_event(self, treeview, data=None):
        if data is not None and data.button == 3:
            paths = self.files_table.get_selected_row_items(0)
            GtkFilesContextMenu(self, data, self.base_dir, paths).show()

    def refresh(self):
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            notification=False
        )
        self.action.append(self.svn.get_items, self.paths, self.svn.STATUSES_FOR_CHECK)
        self.action.append(self.populate_files_table)
        self.action.start()

    @gtk_unsafe
    def populate_files_table(self):
        self.files_table.clear()
        self.items = self.action.get_result(0)
        for item in self.items:
            self.files_table.append([
                item.path, 
                item.simple_content_status(),
                rabbitvcs.util.helper.get_file_extension(item.path)
            ])

    def diff_local(self, path):
        rabbitvcs.util.helper.launch_diff_tool(path)

    def on_context_menu_command_finished(self):
        self.refresh()

class SVNCheckRemoteModifications(GtkContextMenuCaller):
    def __init__(self, caller, vcs, paths, base_dir):
        self.caller = caller
        self.vcs = vcs
        self.svn = vcs.svn()
        self.items = None
        self.paths = paths
        self.base_dir = base_dir

        self.files_table = rabbitvcs.ui.widget.Table(
            self.caller.get_widget("remote_files_table"), 
            [gobject.TYPE_STRING, gobject.TYPE_STRING, 
                gobject.TYPE_STRING, gobject.TYPE_STRING, 
                gobject.TYPE_STRING, gobject.TYPE_STRING], 
            [_("Path"), _("Extension"), 
                _("Text Status"), _("Property Status"), 
                _("Revision"), _("Author")],
            filters=[{
                "callback": rabbitvcs.ui.widget.path_filter,
                "user_data": {
                    "base_dir": base_dir,
                    "column": 0
                }
            }],
            callbacks={
                "row-activated":  self.on_files_table_row_activated,
                "mouse-event":   self.on_files_table_mouse_event
            }
        )

    def on_files_table_row_activated(self, treeview, event, col):
        paths = self.files_table.get_selected_row_items(0)
        self.diff_remote(paths[0])

    def on_files_table_mouse_event(self, treeview, data=None):
        if data is not None and data.button == 3:
            paths = self.files_table.get_selected_row_items(0)
            CheckRemoteModsContextMenu(self, data, self.base_dir, self.vcs, paths).show()

    def refresh(self):
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            notification=False
        )
        self.action.append(self.svn.get_remote_updates, self.paths)
        self.action.append(self.populate_files_table)
        self.action.start()

    @gtk_unsafe
    def populate_files_table(self):
        self.files_table.clear()
        self.items = self.action.get_result(0)
        for item in self.items:
            revision = -1
            author = ""

            if item.revision is not None:
                revision = item.revision
            if item.author is not None:
                author = item.author

            self.files_table.append([
                item.path, 
                rabbitvcs.util.helper.get_file_extension(item.path),
                item.remote_content,
                item.remote_metadata,
                str(revision),
                author
            ])

    def diff_remote(self, path):
        from rabbitvcs.ui.diff import SVNDiff
        
        path_local = path
        path_remote = self.svn.get_repo_url(path_local)
        
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            notification=False
        )
        self.action.append(
            SVNDiff,
            path_local, 
            None, 
            path_remote,
            "HEAD"
        )
        self.action.start()

    def on_context_menu_command_finished(self):
        self.refresh()

class MenuViewDiff(MenuItem):
    identifier = "RabbitVCS::View_Diff"
    label = _("View unified diff")
    icon = "rabbitvcs-diff"

class MenuCompare(MenuItem):
    identifier = "RabbitVCS::Compare"
    label = _("Compare side by side")
    icon = "rabbitvcs-compare"

class CheckRemoteModsContextMenuConditions(GtkFilesContextMenuConditions):
    def __init__(self, vcs, paths=[]):
        GtkFilesContextMenuConditions.__init__(self, vcs, paths)

    def update(self, data=None):
        return True

    def view_diff(self, data=None):
        return (self.path_dict["exists"]
            and self.path_dict["length"] == 1)

    def compare(self, data=None):
        return (self.path_dict["exists"]
            and self.path_dict["length"] == 1)

class CheckRemoteModsContextMenuCallbacks:
    def __init__(self, caller, base_dir, vcs, paths=[]):
        self.caller = caller
        self.base_dir = base_dir
        self.vcs = vcs
        self.svn = self.vcs.svn()
        self.paths = paths

    def update(self, data1=None, data2=None):
        proc = rabbitvcs.util.helper.launch_ui_window(
            "update", 
            self.paths
        )
        self.caller.rescan_after_process_exit(proc, self.paths)

    def view_diff(self, data1=None, data2=None):
        self.caller.diff_remote(self.paths[0])

    def compare(self, data1=None, data2=None):
        from rabbitvcs.ui.diff import SVNDiff
        
        path_local = self.paths[0]
        path_remote = self.svn.get_repo_url(path_local)
        
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            notification=False
        )
        self.action.append(
            SVNDiff,
            path_local, 
            None, 
            path_remote,
            "HEAD",
            sidebyside=True
        )
        self.action.start()

class CheckRemoteModsContextMenu:
    def __init__(self, caller, event, base_dir, vcs, paths=[]):
        
        self.caller = caller
        self.event = event
        self.paths = paths
        self.base_dir = base_dir
        self.vcs = vcs
        
        self.conditions = CheckRemoteModsContextMenuConditions(self.vcs, paths)
        self.callbacks = CheckRemoteModsContextMenuCallbacks(
            self.caller, 
            self.base_dir,
            self.vcs, 
            paths
        )
        
        self.structure = [
            (MenuViewDiff, None),
            (MenuCompare, None),
            (MenuSeparator, None),
            (MenuUpdate, None)
        ]

    def show(self):
        if len(self.paths) == 0:
            return

        context_menu = GtkContextMenu(self.structure, self.conditions, self.callbacks)
        context_menu.show(self.event)

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNCheckForModifications
}

def checkmods_factory(paths, base_dir):
    guess = rabbitvcs.vcs.guess(paths[0])
    return classes_map[guess["vcs"]](paths, base_dir)

if __name__ == "__main__":
    from rabbitvcs.ui import main, BASEDIR_OPT
    (options, paths) = main(
        [BASEDIR_OPT],
        usage="Usage: rabbitvcs checkmods [url_or_path]"
    )

    window = checkmods_factory(paths, options.base_dir)
    window.register_gtk_quit()
    gtk.main()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os
import thread
from time import sleep

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView
from rabbitvcs.util.contextmenu import GtkFilesContextMenu, GtkContextMenuCaller
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.ui.action
import rabbitvcs.util.helper
import rabbitvcs.vcs
from rabbitvcs.util.log import Log
from rabbitvcs.vcs.status import Status

log = Log("rabbitvcs.ui.revert")

from rabbitvcs import gettext
_ = gettext.gettext

gtk.gdk.threads_init()


import rabbitvcs.vcs

log = Log("rabbitvcs.ui.revert")

from rabbitvcs import gettext
_ = gettext.gettext

class Revert(InterfaceView, GtkContextMenuCaller):

    TOGGLE_ALL = True

    def __init__(self, paths, base_dir=None):
        InterfaceView.__init__(self, "revert", "Revert")

        self.paths = paths
        self.base_dir = base_dir
        self.last_row_clicked = None
        self.vcs = rabbitvcs.vcs.VCS()
        self.items = []

        self.statuses = self.vcs.statuses_for_revert(paths)
        self.files_table = rabbitvcs.ui.widget.Table(
            self.get_widget("files_table"),
            [gobject.TYPE_BOOLEAN, rabbitvcs.ui.widget.TYPE_PATH,
                gobject.TYPE_STRING],
            [rabbitvcs.ui.widget.TOGGLE_BUTTON, _("Path"), _("Extension")],
            filters=[{
                "callback": rabbitvcs.ui.widget.path_filter,
                "user_data": {
                    "base_dir": base_dir,
                    "column": 1
                }
            }],
            callbacks={
                "row-activated":  self.on_files_table_row_activated,
                "mouse-event":   self.on_files_table_mouse_event,
                "key-event":     self.on_files_table_key_event
            }
        )

        self.initialize_items()

    def populate_files_table(self):
        self.files_table.clear()
        for item in self.items:
            self.files_table.append([
                True, 
                item.path, 
                rabbitvcs.util.helper.get_file_extension(item.path),
                item.simple_content_status(),
                item.simple_metadata_status()
            ])
                    
    def on_ok_clicked(self, widget):
        return True

    def load(self):
        gtk.gdk.threads_enter()
        self.get_widget("status").set_text(_("Loading..."))
        self.items = self.vcs.get_items(self.paths, self.statuses)

        self.populate_files_table()
        self.get_widget("status").set_text(_("Found %d item(s)") % len(self.items))
        gtk.gdk.threads_leave()

    def populate_files_table(self):
        self.files_table.clear()
        for item in self.items:
            self.files_table.append([
                True,
                item.path,
                rabbitvcs.util.helper.get_file_extension(item.path)
            ])

    # Overrides the GtkContextMenuCaller method
    def on_context_menu_command_finished(self):
        self.initialize_items()

    def initialize_items(self):
        """
        Initializes the activated cache and loads the file items in a new thread
        """

        try:
            thread.start_new_thread(self.load, ())
        except Exception, e:
            log.exception(e)

    def on_select_all_toggled(self, widget):
        self.TOGGLE_ALL = not self.TOGGLE_ALL
        for row in self.files_table.get_items():
            row[0] = self.TOGGLE_ALL

    def on_files_table_row_activated(self, treeview, event, col):
        paths = self.files_table.get_selected_row_items(1)
        rabbitvcs.util.helper.launch_diff_tool(*paths)

    def on_files_table_key_event(self, treeview, data=None):
        if gtk.gdk.keyval_name(data.keyval) == "Delete":
            self.delete_items(treeview, data)

    def on_files_table_mouse_event(self, treeview, data=None):
        if data is not None and data.button == 3:
            self.show_files_table_popup_menu(treeview, data)

    def show_files_table_popup_menu(self, treeview, data):
        paths = self.files_table.get_selected_row_items(1)
        GtkFilesContextMenu(self, data, self.base_dir, paths).show()


class SVNRevert(Revert):
    def __init__(self, paths, base_dir=None):
        Revert.__init__(self, paths, base_dir)
        
        self.svn = self.vcs.svn()

    def on_ok_clicked(self, widget):
        items = self.files_table.get_activated_rows(1)
        if not items:
            self.close()
            return
        self.hide()

        self.action = rabbitvcs.ui.action.SVNAction(
            self.vcs.svn(),
            register_gtk_quit=self.gtk_quit_is_set()
        )
        
        self.action.append(self.action.set_header, _("Revert"))
        self.action.append(self.action.set_status, _("Running Revert Command..."))
        self.action.append(self.vcs.svn().revert, items, recurse=True)
        self.action.append(self.action.set_status, _("Completed Revert"))
        self.action.append(self.action.finish)
        self.action.start()


class GitRevert(Revert):
    def __init__(self, paths, base_dir=None):
        Revert.__init__(self, paths, base_dir)
        
        self.git = self.vcs.git(self.paths[0])

    def on_ok_clicked(self, widget):
        items = self.files_table.get_activated_rows(1)
        if not items:
            self.close()
            return
        self.hide()

        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        
        self.action.append(self.action.set_header, _("Revert"))
        self.action.append(self.action.set_status, _("Running Revert Command..."))
        self.action.append(self.git.checkout, items)
        self.action.append(self.action.set_status, _("Completed Revert"))
        self.action.append(self.action.finish)
        self.action.start()

class SVNRevertQuiet:
    def __init__(self, paths, base_dir=None):
        self.vcs = rabbitvcs.vcs.VCS()
        self.action = rabbitvcs.ui.action.SVNAction(
            self.vcs.svn(),
            run_in_thread=False
        )
        
        self.action.append(self.vcs.svn().revert, paths)
        self.action.run()

class GitRevertQuiet:
    def __init__(self, paths, base_dir=None):
        self.vcs = rabbitvcs.vcs.VCS()
        self.git = self.vcs.git(paths[0])
        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            run_in_thread=False
        )
        
        self.action.append(self.git.checkout, paths)
        self.action.run()

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNRevert, 
    rabbitvcs.vcs.VCS_GIT: GitRevert
}

quiet_classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNRevertQuiet, 
    rabbitvcs.vcs.VCS_GIT: GitRevertQuiet
}

def revert_factory(classes_map, paths, base_dir=None):
    guess = rabbitvcs.vcs.guess(paths[0])
    return classes_map[guess["vcs"]](paths, base_dir)

if __name__ == "__main__":
    from rabbitvcs.ui import main, BASEDIR_OPT, QUIET_OPT
    (options, paths) = main(
        [BASEDIR_OPT, QUIET_OPT],
        usage="Usage: rabbitvcs revert [path1] [path2] ..."
    )

    if options.quiet:
        revert_factory(quiet_classes_map, paths)
    else:
        window = revert_factory(classes_map, paths, options.base_dir)
        window.register_gtk_quit()
        gtk.main()



#
# Copyright (C) 2009 Jason Heeris <jason.heeris@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

"""
Very simple status checking class. Useful when you can't get any of the others
to work, or you need to prototype things. 
"""

import rabbitvcs.vcs
import rabbitvcs.vcs.status

import simplejson

from rabbitvcs import gettext
_ = gettext.gettext

from rabbitvcs.util.log import Log
log = Log("rabbitvcs.services.statuschecker")

class StatusChecker:
    """ A class for performing status checks. """
    
    # All subclasses should override this! This is to be displayed in the
    # settings dialog
    CHECKER_NAME = _("Simple status checker")
    
    def __init__(self):
        """ Initialises status checker. Obviously. """
        self.vcs_client = rabbitvcs.vcs.create_vcs_instance()
        self.conditions_dict_cache = {}

    def check_status(self, path, recurse, summary, invalidate):
        """ Performs a status check, blocking until the check is done.
        """
        path_status = self.vcs_client.status(path, summary, invalidate)
        return path_status
    
    def generate_menu_conditions(self, paths, invalidate=False):
        from rabbitvcs.util.contextmenu import MainContextMenuConditions
        
        conditions = MainContextMenuConditions(self.vcs_client, paths)
        return conditions.path_dict
    
    def extra_info(self):
        return None
    
    def get_memory_usage(self):
        """ Returns any additional memory of any subprocesses used by this
        checker. In other words, DO NOT return the memory usage of THIS process! 
        """
        return 0
    
    def quit(self):
        # We will exit when the main process does
        pass



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

"""

Simple decorators (usable in Python >= 2.4).

Decorators should be named as verbs (present or paste tense).

See: 

  - https://linkchecker.svn.sourceforge.net/svnroot/linkchecker/trunk/linkchecker/linkcheck/decorators.py
  - http://wiki.python.org/moin/PythonDecoratorLibrary
  
"""

import time
import warnings
import threading

from rabbitvcs.util.log import Log
log = Log("rabbitvcs.util.decorators")

def update_func_meta(fake_func, real_func):
    """
    Set meta information (eg. __doc__) of fake function to that of the real 
    function.
    
    @rtype: function
    @return Fake function with metadata of the real function.
    """
    
    fake_func.__module__ = real_func.__module__
    fake_func.__name__ = real_func.__name__
    fake_func.__doc__ = real_func.__doc__
    fake_func.__dict__.update(real_func.__dict__)
    
    return fake_func

def deprecated(func):
    """
    A decorator which can be used to mark functions as deprecated.
    It emits a warning when the function is called.
    
    @type   func: function
    @param  func: The function to be designated as deprecated.
    """
    
    def newfunc(*args, **kwargs):
        """
        Print deprecated warning and execute original function.
        """
        warnings.warn("Call to deprecated function %s." % func.__name__,
                      category=DeprecationWarning)
        return func(*args, **kwargs)
        
    return update_func_meta(newfunc, func)
    
def timeit(func):
    """
    This is a decorator which times a function and prints the time it took in
    milliseconds to stdout.
    
    Based on the timeit function from LinkChecker.
    
    @type   func: function
    @param  func: The function to be timed.
    
    """

    def newfunc(*args, **kwargs):
        """Execute function and print execution time."""
        start_time = time.time()
        result = func(*args, **kwargs)
        duration = (time.time() - start_time) * 1000.0
        log.info("%s() took %0.4f milliseconds" % (func.__name__, duration))
        return result
        
    return update_func_meta(newfunc, func)

def disable(func):
    """
    Disable a function.
    
    @type   func: function
    @param  func: The function to be disabled.
    
    """

    def newfunc(*args, **kwargs):
        return
        
    return update_func_meta(newfunc, func)

def gtk_unsafe(func):
    """
    Used to wrap a function that makes calls to GTK from a thread in
    gtk.gdk.threads_enter() and gtk.gdk.threads_leave().
    
    """
    
    import gtk
    
    def newfunc(*args, **kwargs):
        gtk.gdk.threads_enter()
        result = func(*args, **kwargs)
        gtk.gdk.threads_leave()
        return result
        
    return update_func_meta(newfunc, func)

def debug_calls(caller_log, show_caller=False):
    """
    Given a log to write messages to, wrap a function and log its invocation
    and return. Use like:
    
    @debug_calls(my_modules_log)
    def actual_function(...):
        ...
    
    Warning: do not use with DBUS decorated methods, as this will play havoc
    with introspection.
    """
    
    # We need a function within a function to be able to use the log argument.
    def real_debug(func):
        
        def newfunc(*args, **kwargs):
            caller_log.debug("Calling: %s (%s)" %
                                (func.__name__,
                                 threading.currentThread().getName()))
            
            result = func(*args, **kwargs)
            caller_log.debug("Returned: %s (%s)" %
                                (func.__name__,
                                 threading.currentThread().getName()))
            return result
        
        return update_func_meta(newfunc, func)
    
    return real_debug



#
# test/clone.py
#

import os
from shutil import rmtree
from sys import argv
from optparse import OptionParser

from gittyup.client import GittyupClient
from util import touch

parser = OptionParser()
parser.add_option("-c", "--cleanup", action="store_true", default=False)
(options, args) = parser.parse_args(argv)

DIR = "clone"

if options.cleanup:
    rmtree(DIR, ignore_errors=True)

    print "clone.py clean"
else:
    if os.path.isdir(DIR):
        raise SystemExit("This test script has already been run.  Please call this script with --cleanup to start again")

    g = GittyupClient()
    g.clone("git://github.com/adamplumb/sprout.git", DIR)


    print "clone.py pass"
